
import json
import csv
import gensim
from gensim.models import Word2Vec
from gensim.models import FastText
import tensorflow as tf
import pandas as pd
from gensim.models.callbacks import CallbackAny2Vec
pd.options.mode.chained_assignment = None
import numpy as np
import os
from gensim.models import KeyedVectors
#from sklearn.feature_selection import VarianceThreshold

# os.environ["CUDA_VISIBLE_DEVICES"] = '5'
# gpus = tf.config.experimental.list_physical_devices('GPU')
# tf.debugging.set_log_device_placement(True)
# if gpus:
#     try:
#         # Currently, memory growth needs to be the same across GPUs
#         for gpu in gpus:
#             tf.config.experimental.set_memory_growth(gpu, True)
#             # tf.config.experimental.set_visible_devices(gpus[:1], 'GPU')
#
#         logical_gpus = tf.config.experimental.list_logical_devices('GPU')
#         print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
#     except RuntimeError as e:
#         # Memory growth must be set before GPUs have been initialized
#         print(e)


class EpochLogger(CallbackAny2Vec):
    '''Callback to log information about training'''

    def __init__(self):
        self.epoch = 0

    def on_epoch_begin(self, model):
        print("Epoch #{} start".format(self.epoch))

    def on_epoch_end(self, model):
        print("Epoch #{} end".format(self.epoch))
        self.epoch += 1


class WordEmbedding:
    """
    Class for train and save a word embedding. Load pre-trained embedding model and create embedding vectors.
    """
    def __init__(self , model_file= None,
                 overlapping = True,
                 w2v:int = 1 ,
                 sg:int = 1 ,
                 vectordim: int = 100 ,
                 ngram_len: int = 3 ,
                 AA_substitute: str = True,
                 sequence_max_len: int = None,
                 emb_matrix_file: str= None ,
                 windowsize = 5 ,
                 filename = None):
        """
        Construct a new 'WordEmbedding' object.
        :param model_file: Pre-trained embedding model
        :param overlooping: True = Ngram segmentation using overlaping. False = Ngram segmentation using non-overlaping
        :param w2v:Training architecture: 1 for word2vec; otherwise fastText.
        :param sg:Training algorithm: 1 for skip-gram; otherwise CBOW.
        :param vectordim: Dimensionality of the word vectors.
        :param ngram_len: Length of the ngram
        :param AA_substitute: True = remove all of X amino acids and replace all amino acids B, Z, U and O for N, Q, C and Q.
        :param sequence_max_len: Maximum length of the biological sequence
        :param emb_matrix_file: Pre-trained embedding matrix in a .csv file
        :param windowsize: Maximum distance between the current and predicted word within a sentence
        :param filename: Given name to any files created in this class, no matter what extension
        """
        print('WordEmbedding is running..')
        self.model = None
        self.filename = filename
        self.w2v = w2v
        self.sg = sg
        self.vectordim = vectordim
        self.ngram_len = ngram_len
        self.overlapping = overlapping
        self.windowsize = windowsize
        self.AA_substitute = AA_substitute
        self.sequence_max_len = sequence_max_len
        if model_file:
            self.model_file = True
            self.model = self.load_wv_model(model_file)
        else:
            self.model_file = False
        if emb_matrix_file:
            self.emb_matrix_file = True
            self.emb_matrix = self.load_wv_csv_matrix(emb_matrix_file)
            self.ngramlist = self.create_ngramlist(emb_matrix_file)
        else:
            self.ngramlist = self.create_ngramlist()
            self.emb_matrix_file = False
        self.embedding_vectores = None

    def get_vectores(self):
        x=self.embedding_vectores
        return x

    def get_model(self):
        x=self.model
        return x

    def get_ngramlist(self):
        x = self.ngramlist
        return x

    def get_emb_matrix(self):
        return self.emb_matrix

    def get_rows(self, doc):
        csvreader = csv.reader(doc)
        rows = []
        for row in csvreader:
            rows.append(row)
        return rows


    def aa_substitution(self, sequence):
        """
        Remove all of X amino acids and replace all amino acids B, Z, U and O for N, Q, C and Q.
        """
        sequence.replace('X', '')
        sequence.replace('B', 'N')  # asparagine N / aspartic acid  D - asx - B
        sequence.replace('Z', 'Q')  # glutamine Q / glutamic acid  E - glx - Z
        sequence.replace('U',
                         'C')  # selenocisteina, the closest is the cisteine. but it is a different aminoacid . take care.
        sequence.replace('O', 'K')  # Pyrrolysine to lysine
        return sequence

    def get_list_seqs(self, file, AA_substitute:str = None):
        """
        Get a list of all amino acids sequences from a given file
        :param file: path of the file
        :param AA_substitute: True=Remove all of X amino acids and replace all amino acids B, Z, U and O for N, Q, C and Q.
        :return: list of sequences
        :rtype: list of strings
        """
        #TO-DO meter este
        if AA_substitute != None:
            self.AA_substitute = AA_substitute

        records = self.get_rows(file)
        seqs = []
        #AAList = 'ACDEFGHIKLMNPQRSTVWYUOBXZJ'
        for k in records:
            sequence = k[4]
            if self.AA_substitute == True:
                sequence = self.aa_substitution(sequence)
                seqs.append(sequence)
            else:
                seqs.append(sequence)
        seqs.remove(seqs[0])
        return seqs
####################
    def split_ngrams_non_overlapping(self, seq: str, ngramlen: int = None):
        """
        Split the biological sequence using the non-overlapping method
        :param seq: biological sequence
        :param ngramlen: lenght of the ngram
        :return: list
        """
        #TODO isto só está para ngram = 3
        if ngramlen != None:
            self.ngram_len = ngramlen
        a, b, c = zip(*[iter(seq)] * self.ngram_len), zip(*[iter(seq[1:])] * self.ngram_len), zip(*[iter(seq[2:])] * self.ngram_len)
        str_ngrams = []
        for ngrams in [a, b, c]:
            x = []
            for ngram in ngrams:
                x.append("".join(ngram))
            str_ngrams.append(x)
        return str_ngrams

    def split_ngrams_overlapping(self, seq:list, ngramlen: int = None):
        """
        Split the biological sequence using the overlapping method.
        :param seq: biological sequence
        :param ngramlen: lenght of the ngram
        :return: list of the sequence
        :rtype: list of strings
        """
        # retorna uma lista
        if ngramlen != None:
            self.ngram_len = ngramlen
        chunks = [seq[i:i + self.ngram_len] for i in range(0, len(seq) - self.ngram_len+1, 1)]
        return chunks

    def tokenize_sequences(self, sequences, ngramlen: int = None, overlapping: bool = None):
        """
        Tokenize the biological sequences using the overlapping or non-overlapping methods.
        :param sequences: list of sequences or string of a sequence
        :param ngramlen: lenght of the ngram
        :param overlapping: overlapping method- True: overlapping ; False: non-overlapping
        :return: list of ngrams
        """
        if overlapping != None:
            self.overlapping = overlapping
        if ngramlen != None:
            self.ngram_len = ngramlen

        # print('Tokenizing')
        if self.overlapping == True:
            if type(sequences) == str:
                list_of_ngrams = self.split_ngrams_overlapping(sequences, self.ngram_len)
                return list_of_ngrams
            else:
                seqs_tokenized = []
                for x in sequences:
                    list_of_ngrams = self.split_ngrams_overlapping(x, self.ngram_len)
                    seqs_tokenized.append(list_of_ngrams)
                return seqs_tokenized

        #rv
        elif self.overlapping == False:
            if type(sequences) == str:
                list_of_ngrams = self.split_ngrams_non_overlapping(sequences, self.ngram_len)
                return list_of_ngrams
            else:

                seqs_tokenized = []
                for x in sequences:
                    list_of_ngrams = self.split_ngrams_overlapping(x, self.ngram_len)
                    seqs_tokenized.append(list_of_ngrams)
                return seqs_tokenized


    def sequence_preparation(self,
                             filename: str,
                             list_of_sequences: list,
                             aa_subs: bool = None,
                             overlapping: str=None,
                             ngramlen: int = None):
        """
        Sequence preparation of a given file or list of sequences, tokenizing using different ngrams segmentation methods.
        :param filename: path of the dataset with the sequences
        :param list_of_sequences: list of the biological sequences
        :param aa_subs: True=Remove all of X amino acids and replace all amino acids B, Z, U and O for N, Q, C and Q.
        :param overlapping: overlapping method- True: overlapping ; False: non-overlapping
        :param ngramlen: lenght of the ngram
        :return: list of sequences tokenized
        """
        if aa_subs != None:
            self.AA_substitute = aa_subs
        if overlapping != None:
            self.overlapping = overlapping
        if ngramlen != None:
            self.ngram_len = ngramlen

        if filename is not None:
            file = open(filename)
            list1 = self.get_list_seqs(file, AA_substitute=self.AA_substitute)
        elif list_of_sequences is not None:

            if self.AA_substitute == True:
                list1 = []
                for i in list_of_sequences:
                    list1.append(self.aa_substitution(i))
            else:
                list1 = list_of_sequences
        else:
            print('You should provide a list of sequences!')
            return
        seqs_tokenized = self.tokenize_sequences(list1, ngramlen=self.ngram_len, overlapping=self.overlapping)


        return seqs_tokenized

##################################

    def create_ngramlist(self, matrix = None):
        """
        Create the list of vocabulary, namely the list of all ngrams.
        Obtained through a pre trained matrix or through a pre-created list of all the possible combinations of AA ngrams.
        :param matrix: embedding matrix
        :return: list of all ngrams.
        """
        if matrix:
            x = [i for i in self.emb_matrix.keys()]
        else:
            a_file = open('n_gram_list_only20.json', "r")
            output_dict = json.load(a_file)
            x = output_dict.get(str(self.ngram_len) + '-gram')
            a_file.close()
        return x

    def get_ngramlist_from_newseq(self, sequences: list, ngramlen: int = None, track=True):
        """
        Return a list of all ngrams present in a list of tokenized sequences.
        This method is computationally heavy so it is not advised to use.
        Only use this method instead of the method "create_ngramlist" if you expect a different word
        that is not present in the vocabulary of the pre-created list (with 20 amino acids)
        :param sequences: list of tokenized sequences
        :param ngramlen: length of the ngram
        :param track: if True print the number of sequences scanned.
        :return: set list of ngrams
        """

        print('start geting ngram list')
        if ngramlen != None:
            self.ngram_len = ngramlen
        ngramlist = []
        n = 1
        for sublist in sequences:
            if track is True:
                print('loop:', n, '/', len(sequences))
            else:
                pass
            for x in sublist:
                if x not in ngramlist and len(x) == self.ngram_len:
                    ngramlist.append(x)
            n += 1

        self.ngramlist = ngramlist
        return ngramlist

#######################

    def word2vec_training(self, sequences: list, ngram_list: list =  None, sg:int= None, vectordim: int = None,
                          filename: str = None , windowsize = None , epoch = 10):
        """
        Train, save and set a new word embedding model using word2vec algorithm.
        :param sequences: list of biological sequences.
        :param ngram_list: vocabulary - set list of ngrams.
        :param sg:Training algorithm: 1 for skip-gram; otherwise CBOW.
        :param vector_size:Dimensionality of the word vectors.
        :param filename: Given name to any files created in this class, no matter what extension.
        :param windowsize: Maximum distance between the current and predicted word within a sentence.
        :param epoch:Number of iterations (epochs) over the corpus.
        :return: word embedding vectors.
        """

        print ('Word2Vec..')
        if sg != None:
            self.sg = sg
        if vectordim != None:
            self.vectordim = vectordim
        if ngram_list != None:
            self.ngramlist = ngram_list
        if windowsize != None:
            self.windowsize = windowsize


        callback1 = EpochLogger()
        w2v = Word2Vec(
            sequences,
            window=self.windowsize,
            vector_size=self.vectordim,
            sg=self.sg,
            min_count=1,
            callbacks=[callback1],epochs=epoch)
        word_vectores = w2v.wv
        w2v.wv.save_word2vec_format('word2vec.vector')
        w2v.save(filename)
        print('Model Saved')
        self.model = w2v
        self.embedding_vectores = word_vectores
        return word_vectores

    def load_w2vmodel(self, filename: str):
        """
        Load a pre-trained word2vec model.
        :param filename: Path of the pre-trained model.
        :return: Pre-trained model.
        """
        model = gensim.models.Word2Vec.load(filename)

        self.emb_matrix = KeyedVectors.load(filename, mmap='r')
        self.embedding_vectors = model.wv.vectors
        """if gensim version == 3.8 change use the commented line below"""
        #self.ngramlist = list(model.wv.vocab.keys())
        self.ngramlist = list(model.wv.index_to_key) #gensim == 4.*
        return model

    def fasttext_training(self, sequences: list, ngram_list: list =  None, sg: int = None,
                          filename: str = None , windowsize = None, epoch=10):
        """
        Train, save and set a new word embedding model using fastText algorithm.
        :param sequences: list of biological sequences.
        :param ngram_list: vocabulary - set list of ngrams.
        :param sg:Training algorithm: 1 for skip-gram; otherwise CBOW.
        :param vector_size:Dimensionality of the word vectors.
        :param filename: Given name to any files created in this class, no matter what extension.
        :param windowsize: Maximum distance between the current and predicted word within a sentence.
        :param epoch:Number of iterations (epochs) over the corpus.
        :return: word embedding vectors.
        """
        print('fastText..')
        if sg != None:
            self.sg = sg
        # if vector_size != None:
        #     self.vectordim = vector_size
        if ngram_list != None:
            self.ngramlist = ngram_list
        if windowsize != None:
            self.windowsize = windowsize

        callback1 = EpochLogger()
        ftmodel = FastText(
            sequences,
            window=self.windowsize,
            vector_size=self.vectordim,
            sg=self.sg,
            min_count=1,
            callbacks=[callback1],epochs=epoch)
        word_vectores = ftmodel.wv
        ftmodel.save(filename)
        print ('Model Saved')
        self.model = ftmodel
        self.embedding_vectores = word_vectores
        return  word_vectores

    def load_fasttextmodel(self, filename: str):
        """
        Load a pre-trained fastText model.
        :param filename: Path of the pre-trained model.
        :return: Pre-trained model.
        """
        model = gensim.models.FastText.load(filename)
        self.emb_matrix = KeyedVectors.load(filename, mmap='r')
        self.embedding_vectors = model.wv.vectors
        self.ngramlist = list(model.wv.index_to_key)
        return model
#TODO param**
    def train_wordembedding(self, w2v: int = None, seqs_tokenized: list = None,
                            ngramlist: list = None, sg:int = None ,
                            vectordim = None,
                            ngram_len = None,
                            filename: str = None , save_csv_matrix = True,
                            windowsize = None,epochs = 10):
        """
        Train, save and set a new word embedding model using a specific method.
        :param w2v: Training architecture: 1 for word2vec; otherwise fastText.
        :param seqs_tokenized: Sequences that mus be already tokenized.
        :param ngramlist: vocabulary - set list of ngrams.
        :param sg: Training algorithm: 1 for skip-gram; otherwise CBOW.
        :param filename: Given name to any files created in this method, no matter what extension.
        :param save_csv_matrix: True if you want to save the embedding matrix in a .csv file.
        :param windowsize: Maximum distance between the current and predicted word within a sentence.
        :param epochs: Number of iterations (epochs) over the corpus.
        :return: word embedding vectors.

        """
        print('Start traning word embedding..')

        if sg != None:
            self.sg = sg
        if w2v != None:
            self.w2v = w2v
        if ngramlist != None:
            self.ngramlist = ngramlist
        if windowsize!= None:
            self.windowsize = windowsize
        if vectordim != None:
            self.vectordim = vectordim
        if ngram_len != None:
            self.ngram_len = ngram_len
        if filename:
            self.filename = filename

        path = 'data_processing/'
        # Check whether the specified path exists or not
        isExist = os.path.exists(path)
        if not isExist:
            # Create a new directory because it does not exist
            os.makedirs(path)

        if self.w2v == 1:
            if self.filename:
                filename1 ='data_processing/' + str(self.filename) + '.model'
            else:
                if self.sg == 1:
                    filename1 = 'data_processing/word2vecmodel_skipgram.model'
                elif self.sg == 0:
                    filename1 = 'data_processing/word2vecmodel_cbow.model'

            word_vectores = self.word2vec_training(seqs_tokenized,
                                                   sg=self.sg,
                                                   filename=filename1,
                                                   windowsize = windowsize,
                                                   epoch=epochs)

        elif self.w2v == 0:
            if self.filename:
                filename1 = 'data_processing/' + str(self.filename) + '.model'
            else:
                if self.sg == 1:
                    filename1 = 'data_processing/fasttextmodel_skipgram.model'
                elif self.sg == 0:
                    filename1 = 'data_processing/fasttextmodel_cbow.model'

            word_vectores = self.fasttext_training(seqs_tokenized, sg=self.sg,filename= filename1,windowsize = windowsize,epoch=epochs)

        wordvecDict = {i: '' for i in self.ngramlist}

        for gram in self.ngramlist:
            if gram in word_vectores.key_to_index:
                wordvecDict[gram] = np.ndarray.tolist(word_vectores.get_vector(gram))
            else:
                wordvecDict[gram] = [0] * self.vectordim
        if save_csv_matrix == True:
            self.csv_writer_w_pandas_df(wordvecDict, csv_filename=filename)
            return  wordvecDict , word_vectores
        else:
            dataframe1 = pd.DataFrame.from_dict(wordvecDict, orient='index')
            return dataframe1

    def load_wv_model(self,file: str):
        """
        Load a pre-trained word embedding model.
        :param filename: Path of the pre-trained model.
        :return: Pre-trained model.
        """
        #https://www.shanelynn.ie/word-embeddings-in-python-with-spacy-and-gensim/
        if self.w2v == 1:
            model = self.load_w2vmodel(file)
            print('Model Loaded')
            self.model = model
            return model
        elif self.w2v == 0:
            model = self.load_fasttextmodel(file)
            print('Model Loaded')
            self.model = model
            return model
        else:
            print('Define word embedding tool word2vec or fastText')



    def load_wv_csv_matrix(self,file):
        """
        Load a pre-trained word embedding model from a .csv file.
        :param filename: Path of the pre-trained model.
        :return: Pre-trained model.
        """
        with open(file) as csvfile:
            spamreader = csv.reader(csvfile,delimiter="\t")
            l = []
            for row in spamreader:
                l.append(row)
            l.pop(0)
            matrix_dict = {}
            for i in l:
                a = str(i.pop(0))
                j = []
                for x in i:
                    j.append(float(x))
                matrix_dict[a] = tuple(j)
            print('--MATRIX LOADED--')

            return matrix_dict


    def save_model_matrix(self, data_dict: dict , csv_filename: str = None):
        """
        Save the embedding matrix in a .csv file.
        :param data_dict: Dictionary of the vectors. keys = ngrams ; values=vectors
        :param csv_filename: Given name to the file created.
        """

        if csv_filename == None:
            if self.w2v == 1:
                if csv_filename == None:
                    if self.sg == 1:
                        csv_filename = 'trained_matrix/word2vecmodel_skipgram_matrix.csv'
                    elif self.sg == 0:
                        csv_filename = 'trained_matrix/word2vecmodel_cbow_matrix.csv'
            elif self.w2v == 0:
                if csv_filename == None:
                    if self.sg == 1:
                        csv_filename = 'trained_matrix/fasttextmodel_skipgram_matrix.csv'
                    elif self.sg == 0:
                        csv_filename = 'trained_matrix/fasttextmodel_cbow_matrix.csv'

        myheaders = ['ngram']
        n = 1
        for i in range(self.vectordim):
            myheaders.append('d' + str(n) )
            n += 1
        with open(csv_filename, 'w', newline='') as myfile:
            writer = csv.writer(myfile)
            writer.writerow(myheaders)
            writer.writerows(data_dict.keys())

    def csv_writer_w_pandas_df(self, datadict: dict, csv_filename: str = None):
        """
        Save the embedding matrix in a .csv file using the library pandas.
        :param data_dict: Dictionary of the vectors. keys = ngrams ; values=vectors
        :param csv_filename: Given name to the file created.
        """
        # dict_data: dict with the word vector of every N-gram
        if csv_filename:
            self.filename= csv_filename
        if self.w2v == 1:
            if csv_filename:
                csv_filename = 'data_processing/' + str(self.filename) + '.csv'
            else:
                if self.sg == 1:
                    csv_filename = 'data_processing/word2vecmodel_skipgram_matrix.csv'
                elif self.sg == 0:
                    csv_filename = 'data_processing/word2vecmodel_cbow_matrix.csv'
        elif self.w2v == 0:
            if csv_filename:
                csv_filename = 'data_processing/' + str(self.filename) + '.csv'
            else:
                if self.sg == 1:
                    csv_filename = 'data_processing/fasttextmodel_skipgram_matrix.csv'
                elif self.sg == 0:
                    csv_filename = 'data_processing/fasttextmodel_cbow_matrix.csv'

        dataframe1 = pd.DataFrame.from_dict(datadict, orient='index')
        dataframe1.to_csv(csv_filename, sep='\t')
        print('WordEmbedding Matrix Saved')
####################################
    def get_ngram_vector(self,ngram:str):
        """
        Get ngram vector. Depending on the type of file used (.model or .csv), the way to get the ngram vector
        is different.
        :param ngram: word ngram
        :return: word vector
        """
        if self.model:
            #wordvector = self.emb_matrix.wv.get_vector(ngram, norm=True)
            wordvector = self.emb_matrix.wv.get_vector(ngram)
        elif self.emb_matrix_file:
            wordvector = self.emb_matrix.get(ngram)
        else:
            print('model file not provided!!')
            return
        return wordvector

    def get_sequence_dict_vector(self, sequence):
        """
        Create a dictionary with N-gram as KEY and a N-dim vector as Value.
        :param sequence: biological sequence
        :return: dictionary
        """

        tokenized_sequence = self.tokenize_sequences(sequence)

        dict1 = {}
        for ngram in tokenized_sequence:
            dict1[ngram] = self.get_ngram_vector(ngram)


        return dict1

    def ngram_occurence_counts(self , sequence: str = None):
        """
        Create a dictionary that counts how may times a ngram appears in all sequences given.
        :param sequence: biological sequence
        :return: dictionary
        """
        #faz a contagem dos ngrams
        dict1 = {i: 0 for i in self.ngramlist }
        sequence_toc = self.tokenize_sequences(sequence)

        for i in sequence_toc:
            dict1[i] += 1

        return dict1
##############
    def matrix_method_2(self,sequence = None):
        """
        Create a matrix using the method 2 with shape = (number of sequences, number of words, dimenson of the vector).
        :param sequence: biological sequence
        :return: sequence vector
        """
        dict1 = self.ngram_occurence_counts(sequence)
        matrix = self.emb_matrix

        new_matrix = {}
        for k,v in matrix.items():
            new_matrix[k]=[item * dict1[k] for item in v]

        # my_list = [2, 4, 6]
        # result = [item * 10 for item in my_list]
        # new_matrix = {key: value * int(dict1.get(value)) for key, value in new_matrix.items()}
        return new_matrix

    def vector_method_2(self, sequence = None, ngram_remove_list:list = None):
        """
        Create a vector using the method 2 with shape = (number of sequences , number of words * dimenson of the vector).
        :param sequence: biological sequence
        :return: sequence vector
        """
        matrix = self.matrix_method_2(sequence)
        if ngram_remove_list:
            for k in ngram_remove_list:
                matrix.pop(k, None)


        finalvector = np.array(list(matrix.values())).flatten()

        return finalvector


    def vector_method_3(self, sequence: list = None ):
        """
        Create a vector using the method 3 with shape = (number of sequences , dimenson of the vector).
        :param sequence: biological sequence
        :return: sequence vector
        """

        #multiplica o numero de occurrences do ngram e depois soma
        dict1 = self.ngram_occurence_counts(sequence) #get list de occurrenci de ngrm
        dict2 = self.get_sequence_dict_vector(sequence) #get dict: Key ngram Value Vector da sequencia
        for i in dict2: #multiplica o numero de occurence nos vectores
            multiplier = dict1.get(i)
            listofvaleus = []
            for j in dict2[i]:
                listofvaleus.append(j * int(multiplier))
            dict2[i] = listofvaleus

        dict_values = []
        for i in dict2:
            dict_values.append(dict2.get(i))

        final_vector = [0] * self.vectordim
        for i in dict_values: #soma os vectores todos retornando um vector final de self.vector_dim
            n = 0
            for j in i:
                final_vector[n] += j
                n += 1
        final_vector = np.array(final_vector)
        return final_vector

    def vector_method_1(self, sequence, ngram_remove = False, padding = False):
        """
        Create a vector using the method 1 with shape = (number of sequences , number of words * dimenson of the vector).
        :param sequence: biological sequence
        :param ngram_remove: list of n-grams to remove (can be use when you calculete a varience of the ngram counts)
        :param padding: Padding is an option, when you want to have a fixed input shape.
        :return: sequence vector
        """

        finalvector = []
        tokanized_sequence = self.tokenize_sequences(sequence)

        if ngram_remove:
            for k in ngram_remove:
                if k in ngram_remove:
                    tokanized_sequence.remove(k)
                else: pass

        for ngram in tokanized_sequence:
            finalvector.append(np.array(self.get_ngram_vector(ngram)))

        finalvector = np.array(finalvector)
        if padding == True:
            m = int(self.sequence_max_len) - int(self.ngram_len) + 1
            n = m * int(self.vectordim)
            if len(finalvector) < m:
                dif = m - len(finalvector)
                extravector = np.zeros((dif,self.vectordim))
                finalvector = np.append(finalvector, extravector, axis =0)


        return finalvector

    def convert_seq2vec(self, method:int , sequence: str= None, ngram_remove = None , sequence_max_len = None , padding = False ):
        """
        Convert a biological sequence to a vector through a choosen method of vector creation.
        :param method:
        Method 1:
        Substitute directly the n-grams presented in the sequence by the WE vector. Being K the
        dimension of the word and N the dimension of the WE vector, a sequence of size L will
        be represented by a final vector of (L − k − 1) ∗ N elements. This method preserves the
        spatial information of the location of biological words.
        Method 2:
        In this method, k-mer word frequencies are calculated and multiplied by the corresponding
        WE vectors. A sequence, independent of the size, will be represented by a matrix of
        dimensions Number_of _words ∗ N.
        Method 3:
        All the vectors of Method 2 are summed to reproduce a single vector of dimension N.
        :param sequence: biological sequence
        :param sequence_max_len: Maximum length of the biological sequence
        :param ngram_remove: list of n-grams to remove (can be use when you calculete a varience of the ngram counts)
        :param padding: Padding is an option, when you want to have a fixed input shape.
        :return: sequence vector
        """

        if sequence_max_len:
            self.sequence_max_len = sequence_max_len
            sequence = sequence[0:self.sequence_max_len]
        if method == 1:
            finalvector = self.vector_method_1(sequence, ngram_remove,padding)
            return finalvector

        elif method == 2:
            finalvector = self.vector_method_2(sequence,ngram_remove)
            return finalvector
        elif method == 3:
            finalvector = self.vector_method_3(sequence)
            return finalvector
        else:
            print('define method!')
            return


    def convert_sequences2vec(self, method:int , sequences: list= None, ngram_remove = None , sequence_max_len = None , array_flat= False, padding = False ):
        """
        Convert a biological sequence to a vector through a choosen method of vector creation.
        :param method:
        Method 1:
        Substitute directly the n-grams presented in the sequence by the WE vector. Being K the
        dimension of the word and N the dimension of the WE vector, a sequence of size L will
        be represented by a final vector of (L − k − 1) ∗ N elements. This method preserves the
        spatial information of the location of biological words.
        Method 2:
        In this method, k-mer word frequencies are calculated and multiplied by the corresponding
        WE vectors. A sequence, independent of the size, will be represented by a matrix of
        dimensions Number_of _words ∗ N.
        Method 3:
        All the vectors of Method 2 are summed to reproduce a single vector of dimension N.
        :param sequences:list of biological sequences
        :param array_flat: True if wanted a vector 1 in matrix shape; 3 dimension vector; only used in method 1.
        :param sequence_max_len: Maximum length of the biological sequence
        :param ngram_remove: list of n-grams to remove (can be use when you calculate a variance of the ngram counts)
        :param padding: Padding is an option, when you want to have a fixed input shape.
        :return: sequence vector
        """
        if method == 1:
            vectorized_seqs = []
            for seq in sequences:
                seq_vector = self.convert_seq2vec(method, seq, ngram_remove , sequence_max_len, padding)
                vectorized_seqs.append(seq_vector)

            finalvector = np.array(vectorized_seqs)

            if array_flat == True:
                finalvector = np.reshape(finalvector, (finalvector.shape[0], finalvector.shape[1] * finalvector.shape[2]))
            print('vectors.shape',finalvector.shape)
            return finalvector

        elif method in [2, 3]:
            seqs_vector = []
            for i in sequences:
                vector = self.convert_seq2vec(method=3, sequence=i, ngram_remove=ngram_remove)
                seqs_vector.append(vector)
            seqs_vector = np.array(seqs_vector)
            return seqs_vector
        else:
            print('define method!!')
            return



def main():
    # sequences = ['MLNNILKIATRQSPLAIWQANYVRNQLLSFYPTLLIELVPIVTSGDLILDKPLMKAGGKRLFIKELEQAMLERRADIAVHSMKDITISFPEGIGLAVLCAREDPRDAFISTHYASIDLLPTGAVVGTSSMRRQCQLRERRPDLVMRDLRGNIGTRLEKLDKGEYDALILAAAGLKRLDLAHRIRMIIDPTELLPAVGQGVIGIEYRLEDTHILSILAPLHHSATALRVSAERAMNAKLAGGCQVPIGSYAEIEGDQIWLRALVGSPDGSLIIRSEGRAPLSQAEILGQSIANDLLYRGAESILCRAFQVDS',
    #                 'MMRTIKVGSRRSKLAMTQTKWVIQKLKEINPSFAFEIKEIVTKGDRIVDVTLSKVGGKGLFVKEIEQALLNEEIDMAVHSMKDMPAVLPEGLVIGCIPEREDPRDALISKNRVKLSEMKKGAVIGTSSLRRSAQLLIERPDLTIKWIRGNIDTRLQKLETEDYDAIILAAAGLSRMGWKQDVVTEFLEPERCLPAVGQGALAIECRESDEELLALFSQFTDEYTKRTVLAERAFLNAMEGGCQVPIAGYSVLNGQDEIEMTGLVASPDGKIIFKETVTGNDPEEVGKRCAALMADKGAKDLIDRVKRELDED',
    #                 'MNSETLPAELPATLTIASRESRLAMWQAEHVRDALRKLYPACDVKILGMTTRGDQILDRTLSKVGGKGLFVKELESALADGRADLAVHSLKDVPMELPAGFALAAVMSREDPRDAFVSNDYASLDALPAGAVVGTSSLRREAMLRARYPRLDVRPLRGNLDTRLAKLDRGDYAAIILAAAGLKRLGLAARIRALLDVEDSLPAAGQGALGIEIAAGRADVAAWLAPLHDHATALAVEAERAVSRALGGSCEVPLAAHAVWRGDELHLTGSVSTTDGARVLAARAQSRAATAADALALGRAVSDELERQGARAIVDALVAASAQAQKGGA']
    # w2v = WordEmbedding(w2v=0, sg=1)
    # w2v_sequeces = w2v.sequence_preparation(filename='datasets/ecpred_uniprot_uniref_90.csv')
    # w2v.train_wordembedding(seqs_tokenized=w2v_sequeces)
    # w2v = WordEmbedding(w2v=1, sg=1)
    # w2v_sequeces = w2v.sequence_preparation(filename='datasets/ecpred_uniprot_uniref_90.csv')
    # w2v.train_wordembedding(seqs_tokenized=w2v_sequeces)
    # w2v = WordEmbedding(w2v=0, sg=0)
    # w2v_sequeces = w2v.sequence_preparation(filename='datasets/ecpred_uniprot_uniref_90.csv')
    # w2v.train_wordembedding(seqs_tokenized=w2v_sequeces)
    w2v = WordEmbedding(w2v=1, sg=1 , vectordim=100 , ngram_len= 1 ,windowsize=5)
    w2v_sequeces = w2v.sequence_preparation(filename='/home/igomes/Bumblebee/datasets/ecpred_uniprot_uniref_90_no_negatives.csv' )
    w2v.train_wordembedding(seqs_tokenized=w2v_sequeces , filename = 'w2v_sg_1gram_100dim_10ep_nneg' , epochs= 10)
    w2v = WordEmbedding(w2v=1, sg=1, vectordim=100, ngram_len=2, windowsize=5)
    w2v_sequeces = w2v.sequence_preparation(
        filename='/home/igomes/Bumblebee/datasets/ecpred_uniprot_uniref_90_no_negatives.csv')
    w2v.train_wordembedding(seqs_tokenized=w2v_sequeces, filename='w2v_sg_2gram_100dim_10ep_nneg', epochs=10)


def main2():
    sequence = 'MLNNILKIATRQSPLAIWQANYVRNQLLSFYPTLLIELVPIVTSGDLILDKPLMKAGGKRLFIKELEQAMLERRADIAVHSMKDITISFPEGIGLAVLCIMLNNILKIATRQSPLAIWQANYVRNQLLSFYPTLLIELVPIVTSGDLILDKPLMKAGGKRLFIKELEQAMLERRADIAVHSMKDITISFPEGIGLAVLCIMLNNILKIATRQSPLAIWQANYVRNQLLSFYPTLLIELVPIVTSGDLILDKPLMKAGGKRLFIKELEQAMLERRADIAVHSMKDITISFPEGIGLAVLCIMLNNILKIATRQSPLAIWQANYVRNQLLSFYPTLLIELVPIVTSGDLILDKPLMKAGGKRLFIKELEQAMLERRADIAVHSMKDITISFPEGIGLAVLCIMLNNILKIATRQSPLAIWQANYVRNQLLSFYPTLLIELVPIVTSGDLILDKPLMKAGGKRLFIKELEQAMLERRADIAVHSMKDITISFPEGIGLAVLCIMLNNILKIATRQSPLAIWQANYVRNQLLSFYPTLLIELVPIVTSGDLILDKPLMKAGGKRLFIKELEQAMLERRADIAVHSMKDITISFPEGIGLAVLCIMLNNILKIATRQSPLAIWQANYVRNQLLSFYPTLLIELVPIVTSGDLILDKPLMKAGGKRLFIKELEQAMLERRADIAVHSMKDITISFPEGIGLAVLCIMLNNILKIATRQSPLAIWQANYVRNQLLSFYPTLLIELVPIVTSGDLILDKPLMKAGGKRLFIKELEQAMLERRADIAVHSMKDITISFPEGIGLAVLCIMLNNILKIATRQSPLAIWQANYVRNQLLSFYPTLLIELVPIVTSGDLILDKPLMKAGGKRLFIKELEQAMLERRADIAVHSMKDITISFPEGIGLAVLCIMLNNILKIATRQSPLAIWQANYVRNQLLSFYPTLLIELVPIVTSGDLILDKPLMKAGGKRLFIKELEQAMLERRADIAVHSMKDITISFPEGIGLAVLCI'
    sequences = ['MLNNILKIATRQSPLAIWQANYVRNQLLSFYPTLLIELVPIVTSGDLILDKPLMKAGGKRLFIKELEQAMLERRADIAVHSMKDITISFPEGIGLAVLCAREDPRDAFISTHYASIDLLPTGAVVGTSSMRRQCQLRERRPDLVMRDLRGNIGTRLEKLDKGEYDALILAAAGLKRLDLAHRIRMIIDPTELLPAVGQGVIGIEYRLEDTHILSILAPLHHSATALRVSAERAMNAKLAGGCQVPIGSYAEIEGDQIWLRALVGSPDGSLIIRSEGRAPLSQAEILGQSIANDLLYRGAESILCRAFQVDS',
                     'MMRTIKVGSRRSKLAMTQTKWVIQKLKEINPSFAFEIKEIVTKGDRIVDVTLSKVGGKGLFVKEIEQALLNEEIDMAVHSMKDMPAVLPEGLVIGCIPEREDPRDALISKNRVKLSEMKKGAVIGTSSLRRSAQLLIERPDLTIKWIRGNIDTRLQKLETEDYDAIILAAAGLSRMGWKQDVVTEFLEPERCLPAVGQGALAIECRESDEELLALFSQFTDEYTKRTVLAERAFLNAMEGGCQVPIAGYSVLNGQDEIEMTGLVASPDGKIIFKETVTGNDPEEVGKRCAALMADKGAKDLIDRVKRELDED',
                    'MNSETLPAELPATLTIASRESRLAMWQAEHVRDALRKLYPACDVKILGMTTRGDQILDRTLSKVGGKGLFVKELESALADGRADLAVHSLKDVPMELPAGFALAAVMSREDPRDAFVSNDYASLDALPAGAVVGTSSLRREAMLRARYPRLDVRPLRGNLDTRLAKLDRGDYAAIILAAAGLKRLGLAARIRALLDVEDSLPAAGQGALGIEIAAGRADVAAWLAPLHDHATALAVEAERAVSRALGGSCEVPLAAHAVWRGDELHLTGSVSTTDGARVLAARAQSRAATAADALALGRAVSDELERQGARAIVDALVAASAQAQKGGA']
    #--VARIANCE
    # data1 = pd.read_csv('counts_por_seq.csv')
    # data1 = data1.drop(data1.columns[[0, 1]], axis=1)
    # print('Start Var')
    # sel = VarianceThreshold(0.3)
    # transf = sel.fit_transform(data1)
    # # original dataset without columns
    # ngram_selected = []
    # column_selected = sel.get_support(indices=True)
    # for i in column_selected:
    #     ngram_selected.append(data1.columns[i])
    #print(ngram_selected)
    #WV
    w2vmodel = WordEmbedding(emb_matrix_file='protVec_100d_3grams.csv')
    vector = w2vmodel.convert_seq2vec(method=1, sequence=sequence) #,ngram_remove= ngram_selected)
    print(vector)
    #
    # l = []
    # for seq in sequences:
    #     vector = w2vmodel.convert_seq2vec(method=1, sequence= seq, matrix= True)
    #     l.append(vector)
    #     #print(vector)
    # l = np.array(l)
    #print(l.shape)

def main3():
    wv = WordEmbedding(w2v = 0)
    model = wv.load_fasttextmodel('/home/igomes/Bumblebee/data_processing/fasttest_skip_3gram_100dim_10ep_nneg.model')
    print(model)
    #measure of similarity analogies
    x = model.wv.most_similar('MII')
    print(x)
def create_json_all_AA():
    AAList = 'ACDEFGHIKLMNPQRSTVWY'
    csv_columns = ['1-gram','2-gram','3-gram','4-gram','5-gram']
    dict1 = {'1-gram': [] , '2-gram': [] , '3-gram': [] , '4-gram' : [] , '5-gram': []}
    for aa in AAList:
        dict1['1-gram'].append(aa)
        for aab in AAList:
            dict1['2-gram'].append(aa+aab)
            for aaab in AAList:
                dict1['3-gram'].append(aa + aab + aaab)
                for aaaab in AAList:
                    dict1['4-gram'].append(aa + aab + aaab + aaaab)
                    for aaaaab in AAList:
                        dict1['5-gram'].append(aa + aab + aaab + aaaab + aaaaab)

    # with open('list_of_ngrams1.csv', 'w') as f:
    #     for key in dict1.keys():
    #         f.write("%s,%s\n" % (key, dict1[key]))
    a_file = open("n_gram_list_only20.json", "w")
    json.dump(dict1, a_file)
    a_file.close()
    print (dict1)

if __name__ == "__main__":
    #create_json_all_AA()
    main()





from .word_embedding import WordEmbedding as wv
from sklearn.manifold import TSNE
import numpy as np
import pandas as pd
import pickle
import matplotlib.pyplot as plt
from random import *
_inst = Random()
seed = _inst.seed
random = _inst.random
uniform = _inst.uniform
triangular = _inst.triangular
randint = _inst.randint

NGRAM_PROPERTIES = {
    'A': [71.0788  , 67. , 67. , 7.   , 47. , 6.01 ],
    'C': [103.1448 , 86. , 86. , 8.   , 52. , 5.05 ],
    'D': [115.0886 , 91. , 91. , 18.  , -18., 2.85 ],
    'E': [129.1155 , 109., 109., 17.  , -31., 3.15 ],
    'F': [147.1766 , 135., 135., 4.   , 100., 5.49 ],
    'G': [57.052   , 48. , 48. , 9.   , 0.  , 6.06 ],
    'H': [137.1412 , 118., 118., 13.  , -42., 7.6  ],
    'I': [113.1595 , 124., 124., 2.   , 99. , 6.05 ],
    'K': [128.1742 , 135., 135., 15.  , -23., 9.6  ],
    'L': [113.1595 , 124., 124., 1.   , 100., 6.01 ],
    'M': [131.1986 , 124., 124., 5.   , 74. , 5.74 ],
    'N': [114.1039 , 96. , 96. , 16.  , -41., 5.41 ],
    'P': [97.1167  , 90. , 90. , 11.5 , -46., 6.3  ],
    'Q': [128.1308 , 114., 114., 14.  , 8.  , 5.65 ],
    'R': [156.1876 , 148., 148., 19.  , 41. , 10.76],
    'S': [87.0782  , 73. , 73. , 12.  ,-7.  , 5.68 ],
    'T': [101.1051 , 93. , 93. , 11.  , 13. , 5.6  ],
    'V': [99.1326  , 105., 105., 3.   , 79. , 6.   ],
    'W': [186.2133 , 163., 163., 6.   , 97. , 5.89 ],
    'Y': [163.176  , 141., 141., 10.  , 63. , 5.64 ],
    'X': [142.67295, 134., 134., 4.5  , 88. , 5.45 ],
    'U': [168.064, 0., 0., 0  , 0. , 0. ],
    'O': [255.313, 0., 0., 0  , 0. , 0. ]
}
class ModelTsne:
    """
    ModelTsne train, load and create plots for visualization of the embedding vectors.
    """
    def __init__(self):
        """
        Construct a new 'ModelTsne' object.
        """
        print ('TSNE is running..')

    def make_tsne(self, tokens,
                  filename: str = 'tsne_model.sav',
                  perplexity=500,
                  n_components=2,
                  init='pca',
                  n_iter=1000,
                  random_state=23,
                  verbose=1,
                  learning_rate=1000):
        """
        Trains, save and return a t-SNE model.
        :param tokens: input tokens
        :param filename: name given to the file
        :param perplexity: The perplexity is related to the number of nearest neighbors that is used in other manifold learning algorithms. Larger datasets usually require a larger perplexity. Consider selecting a value between 5 and 50. Different values can result in significantly different results. The perplexity must be less that the number of samples.
        :param n_components: Dimension of the embedded space.
        :param init: initialization of embedding.
        :param n_iter: Maximum number of iterations for the optimization
        :param random_state: Determines the random number generator.
        :param verbose: Verbosity level.
        :param learning_rate:
        :return: t-SNE trained model.
        :rtype: pickle
        """

        tsne_model = TSNE(perplexity=perplexity, n_components=n_components, init=init, n_iter=n_iter, random_state=random_state, verbose=verbose,
                          learning_rate=learning_rate)
        tsne_fit = tsne_model.fit_transform(tokens)

        with open(filename, 'wb') as files:
            pickle.dump(tsne_fit, files)

        return tsne_fit

    def load_tsne(self, filename:str ='trained_models/tsne_model.sav'):
        """
        Load a t-SNE model.
        :param filename: path to file
        :return: pre-trained model.
        :rtype: pickle
        """
        with open(filename, 'rb') as f:
            lr = pickle.load(f)
        print('Model Loaded')
        return lr

    def calculate_property(self, label):
        """
        Calculates the biological properties for a given amino acid.
        properties: Mass, Volume, Van der Waals Volume, Polarity, Hydrophobicity, Charge
        :param label: given amino-acid.
        :return: list of properties values.
        """
        label2 = []
        for aa in label:

            split_to_char = list(aa)
            sum_properties = np.array([0., 0., 0., 0., 0., 0.])
            for char in split_to_char:
                sum_properties += np.array(self.pick_key(char))
            sum_properties /= 3.
            sum_properties = list(sum_properties)
            label2.append(sum_properties)

        return label2

    def pick_key(self, char):
        """
        Replaces the amino acids B,Z,J with a random amino acid present in the rand_dict dictionary.
        :param char: target amino acid
        :return: new amino acid
        :rtype: str
        """
        rand_dict = {1: 'N', 2: 'D', 3: 'E', 4: 'Q', 5: 'L', 6: 'I'}
        try:
            return NGRAM_PROPERTIES[char]
        # return NGRAM_PROPERTIES[char]
        except:
            if char == 'B':
                return NGRAM_PROPERTIES[rand_dict[randint(1, 2)]]
            elif char == 'Z':
                return NGRAM_PROPERTIES[rand_dict[randint(3, 4)]]
            elif char == 'J':
                return NGRAM_PROPERTIES[rand_dict[randint(5, 6)]]

    def visualization(self, X_tsne, label,  filename):
        """
        Visualization of the 6 graphics from the
        attributes calculated for tripeptides as protvec
        :param X_tsne:
        :param label:
        :param filename:
        :return:
        """
        # load final_embedding data
        print('Visualization')
        fig, axarr = plt.subplots(2, 3, figsize=(15, 10))
        # set marker size
        marker_size = 1

        # set scatter
        g1 = axarr[0, 0].scatter(X_tsne[:, 0], X_tsne[:, 1], marker_size, label[:, 0])
        axarr[0, 0].set_title("Mass")
        fig.colorbar(g1, ax=axarr[0, 0])

        g2 = axarr[0, 1].scatter(X_tsne[:, 0], X_tsne[:, 1], marker_size, label[:, 1])
        axarr[0, 1].set_title("Volume")
        fig.colorbar(g2, ax=axarr[0, 1])

        g3 = axarr[0, 2].scatter(X_tsne[:, 0], X_tsne[:, 1], marker_size, label[:, 2])
        axarr[0, 2].set_title("Van der Waals Volume")
        fig.colorbar(g3, ax=axarr[0, 2])

        g4 = axarr[1, 0].scatter(X_tsne[:, 0], X_tsne[:, 1], marker_size, label[:, 3])
        axarr[1, 0].set_title("Polarity")
        fig.colorbar(g4, ax=axarr[1, 0])

        g5 = axarr[1, 1].scatter(X_tsne[:, 0], X_tsne[:, 1], marker_size, label[:, 4])
        axarr[1, 1].set_title("Hydrophobicity")
        fig.colorbar(g5, ax=axarr[1, 1])

        g6 = axarr[1, 2].scatter(X_tsne[:, 0], X_tsne[:, 1], marker_size, label[:, 5])
        axarr[1, 2].set_title("Charge")
        fig.colorbar(g6, ax=axarr[1, 2])


        plt.savefig(filename)
        plt.show()


    def tsne_plot_2D(self, tsne_model= None, model= None, labels= None , filename: str = 'embedding_2d.jpg'):
        """
        Visualization of embedding vectors from a t-SNE model.
        :param tsne_model: pre trained TSNE model
        :param model: input a model to train a TSNE model w/ them
        :param labels: amino acids corresponding to the vectors
        :param filename: name given to the file
        :return: 2D-plot of the TSNE model.
        """

        if tsne_model is None:
            tsne_fit , labels = self.make_tsne(model)
        else:
            tsne_fit = self.load_tsne(tsne_model)

        x = []
        y = []
        for value in tsne_fit:
            x.append(value[0])
            y.append(value[1])

        plt.figure(figsize=(16, 16))

        for i in range(len(x)):
            plt.scatter(x[i], y[i])
            plt.annotate(labels[i],
                         xy=(x[i], y[i]),
                         xytext=(5, 2),
                         textcoords='offset points',
                         ha='right',
                         va='bottom')
        # colors = cm.rainbow(np.linspace(0, 1, 1))
        # plt.scatter(x, y, c=colors, alpha=0.1, label='Coisa')
        # for i, word in enumerate(tsne_fit):
        #     plt.annotate(word, alpha=0.3, xy=(x[i], y[i]), xytext=(5, 2),
        #                  textcoords='offset points', ha='right', va='bottom', size=10)
        plt.legend(loc=4)
        plt.grid(True)
        plt.savefig(filename)
        plt.show()

    def tsne_plot_3D(self, tsne_model= None,
                     model= None,
                     title:str = 'Visualizing Embeddings using t-SNE',
                     label:str = 'Dataset: unknown',
                     filename: str = 'embbed_img_3d.jpg' , a:int = 0.1):
        if tsne_model == None:
            tsne_fit, labels = self.make_tsne(model)
        else:
            tsne_fit = self.load_tsne(tsne_model)

        fig = plt.figure()
        colors = cm.rainbow(np.linspace(0, 1, 1))
        plt.scatter(tsne_fit[:, 0], tsne_fit[:, 1], tsne_fit[:, 2], c=colors, alpha=a,
                    label=label)
        plt.legend(loc=4)
        plt.title(title)
        plt.savefig(filename)
        plt.show()

    def tsne_plot(self, tsne_fit= None, labels= None , filename: str = 'embbed_img_2d.jpg'):
        ''' input:
        tsne_model: pre trained TSNE model
        model: input a model to train a TSNE model w/ them
        '''

        # if tsne_model == None:
        #     tsne_fit , labels = self.make_tsne(model)
        # else:
        #     tsne_fit = self.load_tsne(tsne_model)

        x = []
        y = []
        for value in tsne_fit:
            x.append(value[0])
            y.append(value[1])

        plt.figure(figsize=(16, 16))

        for i in range(len(x)):
            plt.scatter(x[i], y[i])
            plt.annotate(labels[i],
                         xy=(x[i], y[i]),
                         xytext=(5, 2),
                         textcoords='offset points',
                         ha='right',
                         va='bottom')
        plt.savefig(filename)
        plt.show()


    def visualization_2(self, X_tsne , y , filename):
        # load final_embedding data
        print('Visualization')
        fig, axarr = plt.subplots(2, 3, figsize=(15, 10))
        # set marker size
        marker_size = 1

        # set scatter
        g1 = axarr[0, 0].scatter(X_tsne[:, 0], X_tsne[:, 1], marker_size, y)
        axarr[0, 0].set_title("A")
        fig.colorbar(g1)

        plt.savefig(filename)
        plt.show()

    def visualization_3(self, X_1 , y1, X_2 , y2,X_3 , y3,X_4 , y4,X_5 , y5,X_6 , y6 , filename):
        # load final_embedding data
        print('Visualization')
        fig, axarr = plt.subplots(2, 3, figsize=(15, 10))
        # set marker size
        marker_size = 1

        # set scatter
        g1 = axarr[0, 0].scatter(X_1[:, 0], X_1[:, 1], marker_size, y1)
        axarr[0, 0].set_title("w2v Skip-Gram 3gram 100d")
        fig.colorbar(g1, ax=axarr[0, 0])

        g2 = axarr[0, 1].scatter(X_2[:, 0], X_2[:, 1], marker_size, y2)
        axarr[0, 1].set_title("w2v CBow 3gram 100d")
        fig.colorbar(g2, ax=axarr[0, 1])

        g3 = axarr[0, 2].scatter(X_3[:, 0], X_3[:, 1], marker_size, y3)
        axarr[0, 2].set_title("FastText Ski-pgram 3gram 100d")
        fig.colorbar(g3, ax=axarr[0, 2])

        g4 = axarr[1, 0].scatter(X_4[:, 0], X_4[:, 1], marker_size, y4)
        axarr[1, 0].set_title("w2v Skip-Gram 3gram 50d")
        fig.colorbar(g4, ax=axarr[1, 0])

        g5 = axarr[1, 1].scatter(X_5[:, 0], X_5[:, 1], marker_size, y5)
        axarr[1, 1].set_title("w2v Skip-Gram 3gram 20d")
        fig.colorbar(g5, ax=axarr[1, 1])

        g6 = axarr[1, 2].scatter(X_6[:, 0], X_6[:, 1], marker_size, y6)
        axarr[1, 2].set_title("w2v SkipGram with negatives")
        fig.colorbar(g6, ax=axarr[1, 2])
        plt.savefig(filename)
        plt.show()




def main():
    w2v = wv(emb_matrix_file='/home/igomes/Bumblebee/data_processing/w2v_skip_3gram_100dim_10ep_nneg.csv')
    words = w2v.get_emb_matrix()
    result = words.values()
    data = list(result)
    numpyArray = np.array(data)
    result = words.keys()
    data = list(result)
    y = np.array(data)
    tsne = ModelTsne()
    y = tsne.calculate_property(y)
    print(y)
    #tsne.tsne_plot_2D(model='/home/igomes/Bumblebee/trained_models/tsne_model.sav',labels=numpyArray2,filename='embedding_protVec_2d')
    model = tsne.make_tsne(tokens=numpyArray)
    tsne.tsne_plot(tsne_fit=model,labels=y,filename='embedding_bumble_2d')

def main2():
    #w2v = wv(emb_matrix_file='/home/igomes/Bumblebee/protVec_100d_3grams.csv')
    w2v = wv(emb_matrix_file='/home/igomes/Bumblebee/data_processing/w2v_skip_3gram_100dim_10ep_nneg.csv')
    words = w2v.get_emb_matrix()
    words.pop('<unk>', None)
    tsne = ModelTsne()
    vectors = list(words.values())

    numpyArray = np.array(vectors)
    # tsne.tsne_plot_2D(model='/home/igomes/Bumblebee/trained_models/tsne_model.sav',labels=numpyArray2,filename='embedding_protVec_2d')
    model = tsne.make_tsne(tokens=numpyArray)

    ngrams = words.keys()
    y = tsne.calculate_property(ngrams)
    y = np.array(y)
    # set scatter

    tsne.visualization(X_tsne=model, label = y,filename='embedding_visualization_bumble')
def main3():
    #w2v = wv(emb_matrix_file='/home/igomes/Bumblebee/protVec_100d_3grams.csv')
    w2v = wv(emb_matrix_file='/home/igomes/Bumblebee/data_processing/w2v_skip_3gram_100dim_10ep_nneg.csv')
    words = w2v.get_emb_matrix()
    words.pop('<unk>', None)
    #words.pop('WWW', None)
    tsne = ModelTsne()


    vectors = list(words.values())
    numpyArray = np.array(vectors)
    #tsne.tsne_plot_2D(model='/home/igomes/Bumblebee/trained_models/tsne_model.sav',labels=numpyArray2,filename='embedding_protVec_2d')
    model = tsne.make_tsne(tokens=numpyArray)

    ngrams = words.keys()
    df = pd.read_csv('tripeptid.csv', index_col=[0], names=['ngram', 'value'])
    ngrams = pd.DataFrame(ngrams, columns=['ngram'])
    df2 = ngrams.merge(df, how='inner', on='ngram')
    list1 = list(df2['value'])

    tsne.visualization_2(X_tsne=model, y = list1 ,filename='embedding_visualization_docking scores')

def main4():
    df_tri = pd.read_csv('tripeptid.csv', index_col=[0], names=['ngram', 'value'])
    w2v1 = wv(emb_matrix_file='/home/igomes/Bumblebee/data_processing/w2v_skip_3gram_100dim_10ep_nneg.csv')
    w2v2 = wv(emb_matrix_file='/home/igomes/Bumblebee/data_processing/w2v_cbow_3gram_100dim_10ep_nneg.csv')
    w2v3 = wv(emb_matrix_file='/home/igomes/Bumblebee/data_processing/fasttest_skip_3gram_100dim_10ep_nneg.csv')
    w2v4 = wv(emb_matrix_file='/home/igomes/Bumblebee/data_processing/w2v_sg_3gram_50dim_10ep_nneg.csv')
    w2v5 = wv(emb_matrix_file='/home/igomes/Bumblebee/data_processing/w2v_sg_3gram_20dim_10ep_nneg.csv')
    w2v6 = wv(emb_matrix_file='/home/igomes/Bumblebee/data_processing/w2v_skip_3gram_100dim_10ep.csv')
    words = w2v1.get_emb_matrix()
    words.pop('<unk>', None)
    tsne = ModelTsne()
    vectors = list(words.values())
    numpyArray = np.array(vectors)
    model1 = tsne.make_tsne(tokens=numpyArray)
    # print(model1)
    # print(model1.shape)
    ngrams = words.keys()
    ngrams = pd.DataFrame(ngrams, columns=['ngram'])
    # print(ngrams)
    df1 = ngrams.merge(df_tri, how='inner', on='ngram')
    list1 = list(df1['value'])
    # print(df1)
    # print(list1)


    #2

    words = w2v2.get_emb_matrix()
    words.pop('<unk>', None)
    tsne = ModelTsne()
    vectors = list(words.values())
    numpyArray = np.array(vectors)
    model2 = tsne.make_tsne(tokens=numpyArray)
    ngrams = words.keys()
    ngrams = pd.DataFrame(ngrams, columns=['ngram'])
    df2 = ngrams.merge(df_tri, how='inner', on='ngram')
    list2 = list(df2['value'])
    #3
    words = w2v3.get_emb_matrix()
    words.pop('<unk>', None)
    tsne = ModelTsne()
    vectors = list(words.values())
    numpyArray = np.array(vectors)
    model3 = tsne.make_tsne(tokens=numpyArray)
    ngrams = words.keys()
    ngrams = pd.DataFrame(ngrams, columns=['ngram'])
    df3 = ngrams.merge(df_tri, how='inner', on='ngram')
    list3 = list(df3['value'])
    #4
    words = w2v4.get_emb_matrix()
    words.pop('<unk>', None)
    tsne = ModelTsne()
    vectors = list(words.values())
    numpyArray = np.array(vectors)
    model4 = tsne.make_tsne(tokens=numpyArray)
    ngrams = words.keys()
    ngrams = pd.DataFrame(ngrams, columns=['ngram'])
    df4 = ngrams.merge(df_tri, how='inner', on='ngram')
    list4 = list(df4['value'])
    #5
    words = w2v5.get_emb_matrix()
    words.pop('<unk>', None)
    tsne = ModelTsne()
    vectors = list(words.values())
    numpyArray = np.array(vectors)
    model5 = tsne.make_tsne(tokens=numpyArray)
    ngrams = words.keys()
    ngrams = pd.DataFrame(ngrams, columns=['ngram'])
    df5 = ngrams.merge(df_tri, how='inner', on='ngram')
    list5 = list(df5['value'])
    #6
    words = w2v6.get_emb_matrix()
    words.pop('<unk>', None)
    tsne = ModelTsne()
    vectors = list(words.values())
    numpyArray = np.array(vectors)
    model6 = tsne.make_tsne(tokens=numpyArray)
    ngrams = words.keys()
    ngrams = pd.DataFrame(ngrams, columns=['ngram'])
    df6 = ngrams.merge(df_tri, how='inner', on='ngram')
    list6 = list(df6['value'])

    tsne.visualization_3(X_1=model1, y1 = list1
                        ,X_2=model2, y2 = list2,
                         X_3=model3, y3 = list3,
                         X_4=model4, y4 = list4,
                         X_5=model5, y5 = list5,
                         X_6=model6, y6 = list6,
                         filename='embedding_visualization_docking_scores')

if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-
"""
##############################################################################

A function used for a brief comparation of scores between a random forest, gaussian bayes and a SVM with or
without SVC feature selection
Authors:

Date: 06/2019

Email:

##############################################################################
"""
from sklearn import datasets, svm
from sklearn.model_selection import cross_validate, cross_val_score
from sklearn.naive_bayes import GaussianNB
from sklearn.ensemble import RandomForestClassifier


def score_methods(x_original,x_reduced,labels):
    '''

    :param x_original: dataset with features original (no labels)
    :param x_reduced: dataset with feature selection or reduction applied (no labels)
    :param labels: labels of dataset
    :return:
    '''
    svm_mod =svm.SVC(gamma=0.001,C=100)
    gnb = GaussianNB()
    rf=RandomForestClassifier(n_estimators=100)
    scores=cross_val_score(svm_mod,x_original,labels, cv=10)
    print('Score relative to original dataset with SVM:', scores.mean())
    scores_vt=cross_val_score(svm_mod,x_reduced,labels,cv=10)
    print('Score relative to filtered dataset with SVM:', scores_vt.mean())

    scores=cross_val_score(gnb,x_original,labels, cv=10)
    print('Score relative to original dataset with Gaussian:', scores.mean())
    scores_vt=cross_val_score(gnb,x_reduced,labels,cv=10)
    print('Score relative to filtered dataset with Gaussian:', scores_vt.mean())

    scores=cross_val_score(rf,x_original,labels,cv=10)
    print('Score relative to original dataset with random forest:', scores.mean())
    scores_vt=cross_val_score(rf,x_reduced,labels,cv=10)
    print('Score relative to filtered dataset with random forest:', scores_vt.mean())


if __name__=="__main__":
    pass


# -*- coding: utf-8 -*-
"""
############################################################################
File containing a class used for feature selection. The FeatureSelection class aims to select features based on supervised algorithms in order to improve
estimators’ accuracy scores or to boost their performance on very high-dimensional datasets.

Authors: Ana Marta Sequeira

Date:06/2019 altered 12/2020

Email:

##############################################################################
"""

import pandas as pd
from sklearn.feature_selection import GenericUnivariateSelect
from sklearn.feature_selection import f_classif, mutual_info_classif, chi2
# from sklearn.feature_selection import f_regression, mutual_info_regression #for regression problems
from sklearn.feature_selection import SelectFromModel
from sklearn.linear_model import LogisticRegression
from sklearn.svm import LinearSVC
from sklearn.ensemble import ExtraTreesClassifier
from sklearn.linear_model import LogisticRegression
# from scores import score_methods
from sklearn.svm import SVC
from sklearn.feature_selection import RFE, RFECV
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from .ml.utils import timer
import numpy as np


class FeatureSelection:
    """
     The FeatureSelection class aims to select features to improve estimators’ accuracy scores or to boost
     their performance on very high-dimensional datasets.
     It implements sklearn functions
     """

    def _report(self, info, dataframe=False, **kwargs):
        filename = str(self.report_name)
        with open(filename, 'a+') as file:
            if dataframe is True:
                info.to_csv(file, sep='\t', mode='a', **kwargs)
            elif isinstance(info, str):
                file.writelines(info)
            else:
                for l in info:
                    file.writelines('\n{}'.format(l))

    # todo pass the variance threshold to here
    # fix the preprocess to integratae in propythia
    def __init__(self, x_original, target, columns_names, dataset=None, report_name=''):
        """
        init function. When the class is called a dataset containing the features values and a target column must be provided.
        :param dataset: full dataset. optional
        :param x_original: dataset X_data to sklearn_load
        :param target: column with class labels

        """
        self.dataset = dataset
        if columns_names is None:
            if dataset is not None:
                self.columns_names = dataset.columns
            else:
                self.columns_names = list(range(len(self.X_data)))
        else:
            self.columns_names = columns_names

        self.x_original = x_original
        self.X_data = x_original
        self.Y_data = target
        self.column_selected = None
        self.report_name = report_name
        if self.report_name:
            self._report(str(self.report_name))

    def get_x_data(self):
        return self.X_data

    def get_transformed_dataset(self):
        df = pd.DataFrame(self.X_data, columns=self.column_selected)
        df['label'] = self.Y_data
        return df

    @timer
    def run_univariate(self, score_func=mutual_info_classif, mode='percentile', param=50, **params):
        """
        Univariate feature selector, it selects works the best features based on univariate statistical tests.
        It can select the k highest scoring features or a user specified percentage of features.
        Scoring functions for classification problems can be chi2, f_classif or mutual_info_classif

        :param scaler: scaler function to use to datasets before apply univariate tests.
        It can be None or any function supported by SKlearn like StandardScaler()
        :param score_func: function that returns univariate scores and p-values (or only scores for SelectKBest and SelectPercentile)
                    ( for classification: chi2, f_classif, mutual_info_classif
        :param mode: feature selection mode (‘percentile’, ‘k_best’, ‘fpr’, ‘fdr’, ‘fwe’)
        :param param:parameter of corresponding mode
        :return: univariate scores and p-values
        """
        saved_args = locals()

        transformer = GenericUnivariateSelect(score_func, mode=mode, param=param, **params)
        x_fit_univariate = transformer.fit(self.X_data, self.Y_data)
        x_transf_univariate = transformer.transform(self.X_data)
        scores = x_fit_univariate.scores_  # scores of features
        p_values = x_fit_univariate.pvalues_  # p features

        # get columns selected
        column_selected = transformer.get_support(indices=True)  # indexes of selected columns
        self.column_selected = column_selected

        scores_df = self.scores_ranking(scores=scores, df_column_name='scores_ranking', all=False)

        s1 = 'original X dataset shape: {}'.format(self.X_data.shape)
        s2 = 'New X dataset shape: {}'.format(x_transf_univariate.shape)
        s3 = 'number of column selected: {}'.format(column_selected.shape)
        s4 = 'scores: {}'.format(scores_df)
        # s5 = 'column selected: {}'.format(column_selected)
        print('{}\n{}\n{}\n{}'.format(s1, s2, s3, s4))
        # assign the new X
        self.X_data = x_transf_univariate

        if self.report_name:
            self._report([self.run_univariate.__name__, saved_args, s1, s2, s3, s4, s5])
            self._report(scores_df[:100], dataframe=True)

        return transformer, x_fit_univariate, x_transf_univariate, column_selected, scores, scores_df

    # todo retrieves a lot. it gets cofusing and lots of work to do fit in one dataset and transform in other. check.
    #  put the other funtions retrieving in the same way

    @timer
    def run_recursive_feature_elimination(self, cv=None, estimator=SVC(kernel="linear"), n_jobs=None,
                                          step=1, **params):
        """
        Given an external estimator that assigns weights to features (e.g., the coefficients of a linear model), recursive feature elimination (RFE)
        is to select features by recursively considering smaller and smaller sets of features.
        First, the estimator is trained on the initial set of features and the importance of each feature is obtained either
        through a coef_ attribute or through a feature_importances_ attribute.
        Then, the least important features are pruned from current set of features.That procedure is recursively repeated on the pruned set until
        the desired number of features to select is eventually reached.

        RFECV performs RFE in a cross-validation loop to find the optimal number of features.
        :param cv: number of folds in cross validation
        :param estimator: estimator that assign wights to features
        :param n_jobs:
        :param step:If greater than or equal to 1, then step corresponds to the (integer) number of features to remove
        at each iteration. If within (0.0, 1.0), then step corresponds to the percentage (rounded down) of features to
        remove at each iteration
        :return: rfe fit, rfe transformed, original dataset with features selected, columns names and the features ranking
        """
        # todo check if it is ok with small dataset

        saved_args = locals()
        # https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.RFE.html#sklearn.feature_selection.RFE
        # https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.RFECV.html#sklearn.feature_selection.RFECV

        if cv == 1:
            transformer = RFE(estimator, step, **params)
        else:
            transformer = RFECV(estimator, step, cv, n_jobs, **params)

        x_fit_rfe = transformer.fit(self.X_data, self.Y_data)
        x_transf_rfe = transformer.transform(self.X_data)
        # ranking
        features_ranking = transformer.ranking_

        # get columns selected
        column_selected = transformer.get_support(indices=True)  # indexes of selected columns
        self.column_selected = column_selected
        feat_impo_df = self.scores_ranking(scores=features_ranking, df_column_name='features_ranking', all=False)

        s1 = 'original X dataset shape: {}'.format(self.X_data.shape)
        s2 = 'New X dataset shape: {}'.format(x_transf_rfe.shape)
        s3 = 'number of column selected: {}'.format(column_selected.shape)
        s4 = 'features ranking: {}'.format(feat_impo_df)
        # s5 = 'column selected: {}'.format(column_selected)
        print('{}\n{}\n{}\n{}'.format(s1, s2, s3, s4))

        # assign the new X
        self.X_data = x_transf_rfe
        if self.report_name:
            self._report([self.run_recursive_feature_elimination.__name__, saved_args, s1, s2, s3, s4, s5])
            self._report(feat_impo_df[:30], dataframe=True)
        return transformer, x_fit_rfe, x_transf_rfe, column_selected, feat_impo_df

    @timer
    def run_from_model(self, model=LinearSVC(C=0.1, penalty="l1", dual=False), **params):
        """
        SelectFromModel is a meta-transformer that can be used along with any estimator that has a coef_ or feature_importances_ attribute
        after fitting.
        The features are considered unimportant and removed, if the corresponding coef_ or feature_importances_ values are
        below the provided threshold parameter.

        :param model:
                examples:
                ExtraTreesClassifier(n_estimators=50)
                LinearSVC(C=0.01, penalty="l2", dual=False)
                LogisticRegression(C=0.1, penalty="l1", dual=False)

        :return:
        """
        saved_args = locals()

        transformer = model.fit(self.X_data, self.Y_data)
        select_model = SelectFromModel(transformer, prefit=True, **params)
        column_selected = select_model.get_support(indices=True)  # indexes of selected columns
        self.column_selected = column_selected
        x_transf_model = select_model.transform(self.X_data)

        if 'Tree' in str(model):
            feat_impo = transformer.feature_importances_
        else:
            feat_impo = transformer.coef_[0]

        feat_impo_df = self.scores_ranking(scores=feat_impo, df_column_name='features importance', all=False)

        s1 = 'original X dataset shape: {}'.format(self.X_data.shape)
        s2 = 'New X dataset shape: {}'.format(x_transf_model.shape)
        s3 = 'number of column selected: {}'.format(column_selected.shape)
        s4 = 'features importance: {}'.format(feat_impo_df)
        # s5 = 'column selected: {}'.format(column_selected)
        print('{}\n{}\n{}\n{}'.format(s1, s2, s3, s4))

        # assign the new X
        self.X_data = x_transf_model
        if self.report_name:
            self._report([self.run_from_model.__name__, saved_args, s1, s2, s3, s4, s5])
            self._report(feat_impo_df[:30], dataframe=True)
        return select_model, x_transf_model, column_selected, feat_impo, feat_impo_df

    def scores_ranking(self, scores, df_column_name='features ranking', all=False):
        """
        Retrieves a dataframe with features names and scores of importance
        :param scores: list of scores of the features (can be obtained by the function univariate)
        :param column_selected: list containing the indexes of the selected columns (can be obtained by the function univariate)
        :param all: to return all the features and scores or only the selected ones (by default)
        :return: a dataframe containing the names of features and the scores of univariate tests by descending importance
        """
        if all:
            return pd.DataFrame(scores, index=self.columns_names,
                                columns=[df_column_name]).sort_values(by=[df_column_name], ascending=False)
        else:
            # get scores of column selected
            score = []
            for index in self.column_selected:
                score.append(scores[index])
            columns_names_new = np.asarray(self.columns_names)[self.column_selected]

            return pd.DataFrame(score, index=columns_names_new,
                                columns=[df_column_name]).sort_values(by=[df_column_name], ascending=False)



import pandas as pd
import numpy as np
import os

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
# from silence_tensorflow import silence_tensorflow
# silence_tensorflow()
# import logging
# logging.disable(logging.WARNING)
# logging.getLogger("tensorflow").setLevel(logging.FATAL)

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.layers import LSTM, ConvLSTM2D, BatchNormalization

from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Embedding, Bidirectional
from tensorflow.keras.wrappers.scikit_learn import KerasClassifier
from tensorflow.keras import regularizers
from tensorflow.keras.callbacks import ReduceLROnPlateau, EarlyStopping, ModelCheckpoint
from tensorflow.keras.preprocessing.sequence import pad_sequences
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Conv1D
from tensorflow.keras.layers import LSTM
from tensorflow.keras.layers import Flatten
from tensorflow.keras.layers import MaxPool1D
from tensorflow.keras.layers import Dropout
from tensorflow.keras.layers import Input

tf.debugging.set_log_device_placement(True)
strategy = tf.distribute.MirroredStrategy()
# todo por isto como deve ser
# https://towardsdatascience.com/applied-deep-learning-part-4-convolutional-neural-networks-584bc134c1e2
# https://machinelearningmastery.com/cnn-long-short-term-memory-networks/
# https://missinglink.ai/guides/keras/keras-conv1d-working-1d-convolutional-neural-networks-keras/
#
def create_cnn_lstm(input_dim, number_classes,
                    optimizer='Adam',
                    filter_count=(32, 64, 128),
                    padding='same',
                    strides=1,
                    kernel_size=(3,),
                    cnn_activation=None,
                    kernel_initializer='glorot_uniform',
                    dropout_cnn=(0.0,),
                    max_pooling=(True,),
                    pool_size=(2,), strides_pool=1,
                    data_format_pool='channels_first',
                    bilstm=True,
                    lstm_layers=(128, 64),
                    activation='tanh',
                    recurrent_activation='sigmoid',
                    dropout_rate=(0.3,), recurrent_dropout_rate=(0.3,),
                    l1=1e-5, l2=1e-4,
                    dense_layers=(64, 32),
                    dense_activation="relu",
                    dropout_rate_dense=(0.0,),
                    batchnormalization=True,
                    loss_fun='binary_crossentropy', activation_fun='sigmoid'):
    # kernel_initializer='lecun_uniform', activation='relu',

    last_lstm_layer = lstm_layers[-1]
    last_dense_layer = dense_layers[-1]
    # if kernel_size is not list: kernel_size = list([kernel_size]*len(filter_count))
    # if dropout_cnn is not list: dropout_cnn = list([dropout_cnn] * len(filter_count))
    # if max_pooling is not list: max_pooling = list([max_pooling] * len(filter_count))
    # if pool_size is not list: pool_size = list([pool_size] * len(filter_count))
    if len([kernel_size]) == 1:
        kernel_size = list(kernel_size * len(filter_count))
    if len([dropout_cnn]) ==1 : dropout_cnn = list(dropout_cnn * len(filter_count))
    if len([max_pooling]) ==1 : max_pooling = list(max_pooling * len(filter_count))
    if len([pool_size])==1 : pool_size = list(pool_size * len(filter_count))
    if len([dropout_rate]) ==1: dropout_rate = list(dropout_rate * len(lstm_layers))
    # if len([dropout_rate]) == 1:
    #     dropout_rate = list([dropout_rate] * len(lstm_layers))
    if len([recurrent_dropout_rate]) == 1:
        recurrent_dropout_rate = list(recurrent_dropout_rate * len(lstm_layers))
    if len([dropout_rate_dense]) == 1:
        dropout_rate_dense = list(dropout_rate_dense * len(dense_layers))
    if len([batchnormalization]) == 1:
        batchnormalization = list(batchnormalization * len(dense_layers))

    with strategy.scope():
        model = Sequential()
        n_timesteps, n_features = 1, input_dim
        model.add(Input(shape=(n_timesteps,n_features)))
        # model.add(Input(shape=(input_dim,1), dtype='float32', name='main_input'))

        # add convolutional layers
        for cnn in range(len(filter_count)):
            model.add((Conv1D(
                filters=filter_count[cnn],
                kernel_size=kernel_size[cnn],
                strides=strides,
                padding=padding,
                data_format='channels_last',
                activation=cnn_activation,
                dilation_rate=1, use_bias=True,
                kernel_initializer=kernel_initializer, bias_initializer='zeros',
                kernel_regularizer=None, bias_regularizer=None, activity_regularizer=None,
                kernel_constraint=None, bias_constraint=None)))

            max_pool = max_pooling[cnn]
            if max_pool:
                model.add(MaxPool1D(pool_size=pool_size[cnn], strides=strides_pool, padding=padding,
                                    data_format=data_format_pool))
            if dropout_cnn[cnn] > 0:
                model.add(Dropout(dropout_cnn[cnn]))

        # model.add(Flatten())

        # add lstm layers
        if not bilstm:
            # add lstm layers
            for layer in range(len(lstm_layers) - 1):
                model.add(LSTM(units=lstm_layers[layer], return_sequences=True,
                               activation=activation, recurrent_activation=recurrent_activation,
                               dropout=dropout_rate[layer],
                               recurrent_dropout=recurrent_dropout_rate[layer],
                               kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2)))
            # add last lstm layer
            model.add(LSTM(units=last_lstm_layer, return_sequences=False,
                           activation=activation, recurrent_activation=recurrent_activation,
                           dropout=dropout_rate[-1],
                           recurrent_dropout=recurrent_dropout_rate[-1],
                           kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2)))

        elif bilstm:
            # add other lstm layer
            for layer in range(len(lstm_layers) - 1):
                model.add(Bidirectional(LSTM(units=lstm_layers[layer], return_sequences=True,
                                             activation=activation, recurrent_activation=recurrent_activation,
                                             kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2),
                                             dropout=dropout_rate[layer], recurrent_dropout=recurrent_dropout_rate[layer])))
            # add last lstm layer
            model.add(Bidirectional(LSTM(units=last_lstm_layer, return_sequences=False,
                                         activation=activation, recurrent_activation=recurrent_activation,
                                         kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2),
                                         dropout=dropout_rate[-1], recurrent_dropout=recurrent_dropout_rate[-1])))

        # add denses
        for layer in range(len(dense_layers)):
            model.add(Dense(units=dense_layers[layer], activation=dense_activation,
                            kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2)))
            if batchnormalization[layer]:
                model.add(BatchNormalization())
                model.add(Dropout(dropout_rate_dense[layer]))

        # Add Classification Dense, Compile model and make it ready for optimization
        model.add(Dense(number_classes, activation=activation_fun))
        model.compile(loss=loss_fun, optimizer=optimizer, metrics=['accuracy'])
        return model



 # coding: utf-8
#
# #!/usr/bin/env python
#
# """
# Function to calculate fraction of each physico-chemical property present in sequences (separated by comma)
# It receives a sequence and a list containing feature numbers (valid number, 0-24)
#
#     FEATURE NAME                    FEATURE NUMBER
#
#     'Positively charged' --                  0
#     'Negatively charged' --                  1
#     'Neutral charged' --                     2
#     'Polarity' --                            3
#     'Non polarity' --                        4
#     'Aliphaticity' --                        5
#     'Cyclic' --                              6
#     'Aromaticity' --                         7
#     'Acidicity'--                            8
#     'Basicity'--                             9
#     'Neutral (ph)' --                       10
#     'Hydrophobicity' --                     11
#     'Hydrophilicity' --                     12
#     'Neutral' --                            13
#     'Hydroxylic' --                         14
#     'Sulphur content' -                     15
#     'Secondary Structure(Helix)'            16
#     'Secondary Structure(Strands)',         17
#     'Secondary Structure(Coil)',            18
#     'Solvent Accessibilty (Buried)',        19
#     'Solvent Accesibilty(Exposed)',         20
#     'Solvent Accesibilty(Intermediate)',    21
#     'Tiny',                                 22
#     'Small',                                23
#     'Large'                                 24
# Author: Ana Marta Sequeira
#
# email:
#
# date: 05/2019
#
# """
#
#
#
# # Secondary functions, move to next section
# # Single function to calculate 30 physico che
# import sys
# import os
# import pandas as pd
# import numpy as np
# import csv
# import getopt
#
# #Finding physico-chemical property of a vector of polypeptides
#
#
# def init(path=None, index=''):
#     """
#     Read in files. You need to run this (once) before you can
#     access any records. If the files are not within the current directory,
#     you need to specify the correct directory path.
#
#     :param path:
#     :param index:
#     :return:
#
#     """
#     index = str(index)
#     if path is None:
#         for path in [os.path.split(__file__)[0]]:
#             if os.path.exists(os.path.join(path, index)):
#                 break
#         # print('path =', path, file=sys.stderr)
#
#     doc=''
#     if index == 'bonds.csv':
#         doc=(path + '/data/bonds.csv')
#     if index == 'PhysicoChemical.csv':
#         doc=(path + '/data/PhysicoChemical.csv')
#     return doc
#
#
# doc=init(index='PhysicoChemical.csv')
# #Finding physico-chemical property of a vector of polypeptides
# PCP = pd.read_csv(doc, header=None) #Our reference table for properties
#
# headers = ['Positively charged',
#            'Negatively charged',
#            'Neutral charged',
#            'Polarity',
#            'Non polarity',
#            'Aliphaticity',
#            'Cyclic',
#            'Aromaticity',
#            'Acidicity',
#            'Basicity',
#            'Neutral (ph)',
#            'Hydrophobicity',
#            'Hydrophilicity',
#            'Neutral',
#            'Hydroxylic',
#            'Sulphur content',
#            'Secondary Structure(Helix)',
#            'Secondary Structure(Strands)',
#            'Secondary Structure(Coil)',
#            'Solvent Accessibilty (Buried)',
#            'Solvent Accesibilty(Exposed)',
#            'Solvent Accesibilty(Intermediate)',
#            'Tiny',
#            'Small',
#            'Large',
#            'z1',
#            'z2',
#            'z3',
#            'z4',
#            'z5'];
#
#
# def encode(peptide):
#     #letter = {'A','C','D','E','F','G','H','I','K','L','M','N','P','Q','R','S','T','V','W','Y'};
#     l=len(peptide)
#     encoded=np.zeros(l)
#     for i in range(l):
#         if(peptide[i]=='A'):
#             encoded[i] = 0
#         elif(peptide[i]=='C'):
#             encoded[i] = 1
#         elif(peptide[i]=='D'):
#             encoded[i] = 2
#         elif(peptide[i]=='E'):
#             encoded[i] = 3
#         elif(peptide[i]=='F'):
#             encoded[i] = 4
#         elif(peptide[i]=='G'):
#             encoded[i] = 5
#         elif(peptide[i]=='H'):
#             encoded[i] = 6
#         elif(peptide[i]=='I'):
#             encoded[i] = 7
#         elif(peptide[i]=='K'):
#             encoded[i] = 8
#         elif(peptide[i]=='L'):
#             encoded[i] = 9
#         elif(peptide[i]=='M'):
#             encoded[i] = 10
#         elif(peptide[i]=='N'):
#             encoded[i] = 11
#         elif(peptide[i]=='P'):
#             encoded[i] = 12
#         elif(peptide[i]=='Q'):
#             encoded[i] = 13
#         elif(peptide[i]=='R'):
#             encoded[i] = 14
#         elif(peptide[i]=='S'):
#             encoded[i] = 15
#         elif(peptide[i]=='T'):
#             encoded[i] = 16
#         elif(peptide[i]=='V'):
#             encoded[i] = 17
#         elif(peptide[i]=='W'):
#             encoded[i] = 18
#         elif(peptide[i]=='Y'):
#             encoded[i] = 19
#         else:
#             print('Wrong residue!')
#     return encoded
#
#
# def lookup(peptide,featureNum):
#     l=len(peptide)
#     peptide = list(peptide)
#     out=np.zeros(l)
#     peptide_num = encode(peptide)
#
#     for i in range(l):
#         out[i] = PCP[peptide_num[i]][featureNum]
#     return sum(out)
#
#
# def pcp(seq):
#
#     l = len(seq)
#
#     rows = PCP.shape[0] # Number of features in our reference table
#     col = 20  # Denotes the 20 amino acids
#
#     seq=[seq[i].upper() for i in range(l)]
#     sequenceFeature = {}
#
#
#     for i in range(l): # Loop to iterate over each sequence
#         nfeatures = rows
#         for j in range(nfeatures): #Loop to iterate over each feature
#             featureVal = lookup(seq[i],j)
#             if(len(seq[i])!=0):
#                 sequenceFeature[headers[j]]=(round(featureVal/len(seq[i]),3))
#             else:
#                 sequenceFeature[headers[j]]='NaN'
#     print(sequenceFeature)
#     return sequenceFeature
#
# #
# # # In[68]:
# #
# #
# # '''
# #
# # unction Name: phyChem
# # Description: Gives 30 physico-chemical properties of a sequence
# #
# # Function prototype: phyChem(file,mode,m,n)
# #
# # Input:
# # @file: an input csv file with multiple sequences
# # @mode(optional, default = 'all'):
# #     Values possible:
# #         1) (default)'all' : to get features of entire protein
# #         2) 'NT' : to get the features of first n N-Terminal residues
# #         3) 'CT' : to get the features of last n C-Terminal residues
# #         4) 'rest' : to get the features of a sub-sequence from mth position to nth position(both inclusive)
# # @m(optional(mandatory if 'rest' is chosen, default=0): m is the start position of residue
# # @n(optional, default = '0'): n is the number of residues you want from desired terminal or end point (if 'rest' is chosen)
# #
# #
# # Output:
# # A matrix (csv file) of dimension (m x 30) containing sequences as rows and their 30 physico-chemical properties as columns
# # where m = number of sequences (each sequence separated by comma)
# #
# # '''
# #
# #
# # '''
# #
# # Function Name: phyChem
# # Description: Gives 30 physico-chemical properties of a sequence
# #
# # Function prototype: phyChem(file,mode,m,n)
# #
# # Input:
# # @file: an input csv file with multiple sequences
# # @mode(optional, default = 'all'):
# #     Values possible:
# #         1) (default)'all' : to get features of entire protein
# #         2) 'NT' : to get the features of first n N-Terminal residues
# #         3) 'CT' : to get the features of last n C-Terminal residues
# #         4) 'rest' : to get the features of a sub-sequence from mth position to nth position(both inclusive)
# # @m(optional(mandatory if 'rest' is chosen, default=0): m is the start position of residue
# # @n(optional, default = '0'): n is the number of residues you want from desired terminal or end point (if 'rest' is chosen)
# #
# #
# # Output:
# # A matrix (csv file) of dimension (m x 30) containing sequences as rows and their 30 physico-chemical properties as columns
# # where m = number of sequences (each sequence separated by comma)
# #
# # '''
# #
# #
# # def pcp_wp(file,outt,mode='all',m=0,n=0):
# #
# #     if(type(file) == str):
# #         seq = pd.read_csv(file,header=None, sep=',');
# #         seq=seq.T
# #         seq[0].values.tolist()
# #         seq=seq[0];
# #     #elif(type(file)==str && len(file))
# #     else:
# #         seq  = file;
# #
# #     l = len(seq);
# #
# #     newseq = [""]*l; # To store the trimmed sequence
# #     #print('Original Sequence');
# #     #print(seq)
# #
# #
# #     for i in range(0,l):
# #
# #         #if(n<len(seq[i])):
# #         l = len(seq[i]);
# #
# #         if(mode=='NT'):
# #             n=m;
# #             if(n!=0):
# #                 newseq[i] = seq[i][0:n];
# #
# #             elif(n>l):
# #                 print('Warning! Sequence',i,"'s size is less than n. The output table would have NaN for this sequence");
# #
# #             else:
# #                 print('Value of n is mandatory, it cannot be 0')
# #                 break;
# #
# #         elif(mode=='CT'):
# #             n=m;
# #             if(n!=0):
# #                 newseq[i] = seq[i][(len(seq[i])-n):]
# #
# #             elif(n>l):
# #                 print('WARNING: Sequence',i+1,"'s size is less than the value of n given. The output table would have NaN for this sequence");
# #
# #
# #             else:
# #                 print('Value of n is mandatory, it cannot be 0')
# #                 break
# #
# #         elif(mode=='all'):
# #             newseq = seq
# #
# #         elif(mode=='rest'):
# #             if(m==0):
# #                 print('Kindly provide start index for rest, it cannot be 0');
# #                 break
# #
# #             else:
# #                 if(n<=len(seq[i])):
# #                     newseq[i] = seq[i][m:(len(seq[i])-n)]
# #                 '''elif(n>len(seq[i])):
# #                     newseq[i] = seq[i][m-1:len(seq[i])]
# #                     print('WARNING: Since input value of n for sequence',i+1,'is greater than length of the protein, entire sequence starting from m has been considered')'''
# # #
# # #         else:
# # #             print("Wrong Mode. Enter 'NT', 'CT','all' or 'rest'");
# #
# #
# #     output = pcp(newseq,outt);
# #     return output
# #
# #
#
# if __name__ == '__main__':
#     pcp('KRVLLLDNLSDYIKPGMSVEAIQGIIASMKSDYEDRVDDYIIKNAELSKERRDISKKLKVMGE')
#     pcp('VVLSDQGALFEPCSTQDIACLSRATQQFLEKACRGVPEYDIRPIDPLIISSLDVAAYDDIGLIFHFKNLNITGLKNQKISDFRMDTTRKSVLLKTQADLNVVADVVIELSKQSKSFAGVMNIQASIIGGAKYSYDLQDDSKGVKHFEVGQETISCESIGEPAVNLNPELADALLKDPDTTHYRKDYEAHRVSIRQ')


# -*- coding: utf-8 -*-
#!/usr/bin/env python
"""
This function gives the sum of the bond composition for each type of bond
For bond composition four types of bonds are
considered total number of bonds (including aromatic), hydrogen bond, single bond and double
bond. The number of values for each kind of bond is provided as bonds.csv file

The code is based on the package Pfeature :
Pande, Akshara & Patiyal, Sumeet & Lathwal, Anjali & Arora, Chakit & Kaur, Dilraj & Dhall, Anjali & Mishra, Gaurav & Kaur,
Harpreet & Sharma, Neelam & Jain, Shipra & Usmani, Salman & Agrawal, Piyush & Kumar, Rajesh & Kumar, Vinod & Raghava, Gajendra.
(2019). Computing wide range of protein/peptide features from their sequence and structure. 10.1101/599126.


It returns a dictionary form with the values

Authors: Ana Marta Sequeira

Date: 05/2019

Email:


##############################################################################
"""
import pandas as pd
import os
# import sys
def init(path=None, index=''):
    """
    Read in files. You need to run this (once) before you can
    access any records. If the files are not within the current directory,
    you need to specify the correct directory path.

    :param path:
    :param index:
    :return:
    """
    index = str(index)
    if path is None:
        for path in [os.path.split(__file__)[0]]:
            if os.path.exists(os.path.join(path, index)):
                break
        # print('path =', path, file=sys.stderr)

    doc=''
    if index == 'bonds.csv':
        doc=(path + '/data/bonds.csv')
    if index == 'PhysicoChemical.csv':
        doc=(path + '/data/PhysicoChemical.csv')
    return doc


doc=init(index='bonds.csv')
#Finding physico-chemical property of a vector of polypeptides
bonds = pd.read_csv(doc, header=None)


def boc_wp(seq):
    """
    Sum of the bond composition for each type of bond: total number of bonds (including aromatic), hydrogen bond,
    single bond and double

    :param seq: protein sequence
    :return: dictionary with number of total, hydrogen, single and double bonds
    """
    tota = []
    hy = []
    Si = []
    Du = []
    b1 = []
    b2 = []
    b3 = []
    b4 = []
    bb = {}

    df = seq

    for i in range(0,len(df)) :
        tot = 0
        h = 0
        S = 0
        D = 0
        tota.append([i])
        hy.append([i])
        Si.append([i])
        Du.append([i])
        for j in range(0,len(df[i])) :
            temp = df[i][j]
            for k in range(0,len(bonds)) :
                if bonds.iloc[:,0][k] == temp :
                    tot = tot + int(bonds.iloc[:,1][k])
                    h = h + int(bonds.iloc[:,2][k])
                    S = S + int(bonds.iloc[:,3][k])
                    D = D + int(bonds.iloc[:,4][k])
        tota[i].append(tot)
        hy[i].append(h)
        Si[i].append(S)
        Du[i].append(D)
    for m in range(0,len(df)) :
        b1.append(tota[m][1])
        b2.append(hy[m][1])
        b3.append(Si[m][1])
        b4.append(Du[m][1])

    #IF BOND COMPOSITION SEPARATEDLY BY AA RESIDUE JUST TAKE OFF SUM (AND WILL GIVE 4*LEN(SEQ)
    bb["tot"] = sum(b1)
    bb["hydrogen"] = sum(b2)
    bb["single"] = sum(b3)
    bb["double"] = sum(b4)
    
    return bb


if __name__ == '__main__':
    print(boc_wp('MQGNGSALPNASQPVLRGDGARPSWLASALACVLIFTIVVDILGNLLVILSVYRNKKLRN'))
    print(boc_wp('MALPNAVIAAAALSVYRNKKLRN'))
    print(boc_wp('MQGNGSPALLNSRRRRRGDGARPSWLASALACVLIFTIVVDILGNLLVILSVYRNKKLRN'))




import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import logging
logging.getLogger("tensorflow").setLevel(logging.ERROR)
import tensorflow as tf
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
from tensorflow.keras import backend as K
from tensorflow.keras.layers import *

class attention(Layer):
    def __init__(self, **kwargs):
        super(attention, self).__init__(**kwargs)

    def build(self, input_shape):
        self.W = self.add_weight(name="att_weight", shape=(input_shape[-1], 1), initializer="normal")
        self.b = self.add_weight(name="att_bias", shape=(input_shape[1], 1), initializer="zeros")
        super(attention, self).build(input_shape)

    def call(self, x):
        et = K.squeeze(K.tanh(K.dot(x, self.W) + self.b), axis=-1)
        at = K.softmax(et)
        at = K.expand_dims(at, axis=-1)
        output = x * at
        return K.sum(output, axis=1)

    def compute_output_shape(self, input_shape):
        return (input_shape[0], input_shape[-1])

    def get_config(self):
        return super(attention, self).get_config()



import pandas as pd
import numpy as np
import os

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
# from silence_tensorflow import silence_tensorflow
# silence_tensorflow()
# import logging
# logging.disable(logging.WARNING)
# logging.getLogger("tensorflow").setLevel(logging.FATAL)

import tensorflow as tf
from tensorflow import keras

from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Embedding, Bidirectional
from tensorflow.keras.wrappers.scikit_learn import KerasClassifier
from tensorflow.keras import regularizers
from tensorflow.keras.callbacks import ReduceLROnPlateau, EarlyStopping, ModelCheckpoint
from tensorflow.keras.preprocessing.sequence import pad_sequences
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Conv1D, Conv2D
from tensorflow.keras.layers import LSTM
from tensorflow.keras.layers import Flatten
from tensorflow.keras.layers import MaxPool1D, MaxPool2D
from tensorflow.keras.layers import Dropout
from tensorflow.keras.layers import Input

tf.debugging.set_log_device_placement(True)
strategy = tf.distribute.MirroredStrategy()


# todo por isto como deve ser
# https://towardsdatascience.com/applied-deep-learning-part-4-convolutional-neural-networks-584bc134c1e2
# https://missinglink.ai/guides/keras/keras-conv1d-working-1d-convolutional-neural-networks-keras/
# https://towardsdatascience.com/applied-deep-learning-part-4-convolutional-neural-networks-584bc134c1e2
# https://towardsdatascience.com/simple-introduction-to-convolutional-neural-networks-cdf8d3077bac


# 1D networks allow you to use larger filter sizes.
# In a 1D network, a filter of size 7 or 9 contains only 7 or 9 feature vectors.
# Whereas in a 2D CNN, a filter of size 7 will contain 49 feature vectors.
# Can use larger convolution windows with 1D CNNs.
# With a 2D convolution layer, a 3 × 3 convolution window contains 3 × 3 = 9 feature vectors.
# With 1D convolution layer, a window of size 3 contains only 3 feature vectors.
# You can thus easily afford 1D convolution windows of size 7 or 9.

# todo explain really well this list system of mounting a cnn
def create_cnn_1D(input_dim, number_classes,
                  optimizer='Adam',
                  filter_count=(32, 64, 128),  # define number layers
                  padding='same',
                  strides=1,
                  kernel_size=(3,),  # list of kernel sizes per layer. if number will be the same in all numbers
                  cnn_activation='relu',
                  kernel_initializer='glorot_uniform',
                  dropout_cnn=(0.0, 0.2, 0.2),
                  # list of dropout per cnn layer. if number will be the same in all numbers
                  max_pooling=(True,),
                  pool_size=(2,), strides_pool=1,
                  data_format_pool='channels_first',
                  dense_layers=(64, 32),
                  dense_activation="relu",
                  dropout_rate=(0.3,),
                  l1=1e-5, l2=1e-4,
                  loss_fun='binary_crossentropy', activation_fun='sigmoid'):
    # kernel_initializer='lecun_uniform', activation='relu',

    last_dense_layer = dense_layers[-1]
    # todo list !!! they are tuples receive ints in dropouts
    if len(kernel_size) == 1:
        kernel_size = list(kernel_size * len(filter_count))
    if len(dropout_cnn) ==1 : dropout_cnn = list(dropout_cnn * len(filter_count))
    if len(max_pooling) ==1 : max_pooling = list(max_pooling * len(filter_count))
    if len(pool_size)==1 : pool_size = list(pool_size * len(filter_count))
    if len(dropout_rate) ==1: dropout_rate = list(dropout_rate * len(dense_layers))

    with strategy.scope():
        model = Sequential()
        n_timesteps, n_features = 1, input_dim
        model.add(Input(shape=(n_timesteps,n_features)))

        # model.add(layers.Embedding(max_features, 128, input_length=max_len))
        # https://machinelearningmastery.com/cnn-models-for-human-activity-recognition-time-series-classification/
        # add convolutional layers
        for cnn in range(len(filter_count)):
            model.add(Conv1D(
                filters=filter_count[cnn],
                kernel_size=kernel_size[cnn],
                strides=strides,
                padding=padding,
                data_format='channels_last',
                activation=cnn_activation,
                dilation_rate=1, use_bias=True,
                kernel_initializer=kernel_initializer, bias_initializer='zeros',
                kernel_regularizer=None, bias_regularizer=None, activity_regularizer=None,
                kernel_constraint=None, bias_constraint=None))

            max_pool = max_pooling[cnn]
            if max_pool:
                model.add(MaxPool1D(pool_size=pool_size[cnn], strides=strides_pool, padding=padding,
                                    data_format=data_format_pool))

            if dropout_cnn[cnn] > 0:
                model.add(Dropout(dropout_cnn[cnn]))

        # Flatten
        model.add(Flatten())
        # add dense layers
        for layer in range(len(dense_layers)):
            model.add(Dense(units=dense_layers[layer], activation=dense_activation,
                            kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2)))
            model.add(Dropout(dropout_rate[layer]))

        # Add Classification Dense, Compile model and make it ready for optimization
        model.add(Dense(number_classes, activation=activation_fun))
        model.compile(loss=loss_fun, optimizer=optimizer, metrics=['accuracy'])
        return model

# todo explain really well this list system of mounting a cnn
def create_cnn_2D(input_dim, number_classes,
                  optimizer='Adam',
                  filter_count=(32, 64, 128),  # define number layers
                  padding='same',
                  strides=1,
                  kernel_size=((3,3),),  # list of kernel sizes per layer. if number will be the same in all numbers
                  cnn_activation='relu',
                  kernel_initializer='glorot_uniform',
                  dropout_cnn=(0.0, 0.2, 0.2),
                  # list of dropout per cnn layer. if number will be the same in all numbers
                  max_pooling=True,
                  pool_size=((2,2),), strides_pool=1,
                  data_format_pool='channels_last',
                  dense_layers=(64, 32),
                  dense_activation="relu",
                  dropout_rate=0.3,
                  l1=1e-5, l2=1e-4,
                  loss_fun='binary_crossentropy', activation_fun='sigmoid'):
    # kernel_initializer='lecun_uniform', activation='relu',

    last_dense_layer = dense_layers[-1]
    if len(kernel_size) == 1:
        kernel_size = list(kernel_size * len(filter_count))
    if len(dropout_cnn) ==1 : dropout_cnn = list(dropout_cnn * len(filter_count))
    if len([max_pooling]) ==1 : max_pooling = list(max_pooling * len(filter_count))
    if len(pool_size)==1 : pool_size = list(pool_size * len(filter_count))
    if len([dropout_rate]) ==1: dropout_rate = list(dropout_rate * len(dense_layers))

    with strategy.scope():
        model = Sequential()
        # n_timesteps, n_features = 1, input_dim
        # model.add(Input(shape=(n_timesteps,n_features)))
        # input shape = batch size + (image * image, channels)
        model.add(Input(shape=(input_dim)))

        # model.add(Input(shape=(input_dim,), dtype='float32', name='main_input'))

        # add convolutional layers
        for cnn in range(len(filter_count)):
            model.add(Conv2D(
                filters=filter_count[cnn],
                kernel_size=kernel_size[cnn],
                strides=strides,
                padding=padding,
                data_format='channels_last',
                activation=cnn_activation,
                dilation_rate=1, use_bias=True,
                kernel_initializer=kernel_initializer, bias_initializer='zeros',
                kernel_regularizer=None, bias_regularizer=None, activity_regularizer=None,
                kernel_constraint=None, bias_constraint=None))

            max_pool = max_pooling[cnn]
            if max_pool:
                model.add(MaxPool2D(pool_size=pool_size[cnn], strides=strides_pool, padding=padding,
                                    data_format=data_format_pool))
            if dropout_cnn[cnn] > 0:
                model.add(Dropout(dropout_cnn[cnn]))

        # Flatten
        model.add(Flatten())
        # add dense layers
        for layer in range(len(dense_layers)):
            model.add(Dense(units=dense_layers[layer], activation=dense_activation,
                            kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2)))
            model.add(Dropout(dropout_rate[layer]))


        # Add Classification Dense, Compile model and make it ready for optimization
        # model binary
        model.add(Dense(number_classes, activation=activation_fun))
        model.compile(loss=loss_fun, optimizer=optimizer, metrics=['accuracy'])
        return model

# important parameters to test
# filter size
# filter count this is the most variable parameter,
# it’s a power of two anywhere between 32 and 1024. Using more filters results in a more powerful model,
# but we risk overfitting due to increased parameter count. Usually we start with a small number of
# filters at the initial layers, and progressively increase the count as we go deeper into the network.

# todo if this one works do one for 2D (matrixes)

# # https://github.com/BadrulAlom/Protein-Function-CNN-Model/blob/master/ProteinFunction_CNNModel/code/BioScript.py
# class ProteinSequencingCNNModel:
#     def __init__(self):
#         self.input_Height = 1  # Set to 0 to start with
#         self.input_Width = 1  # Set to 0 to start with
#         self.input_Channels = 1  # (e.g. 3 for RGB)
#         self._input_TrainData = []
#         self._input_TrainLabels = []
#         self._input_TestData = []
#         self._input_PredData = []
#         self._input_TestLabels = []
#
#         self.output_FilePath = ""
#         self.output_Filename = "ProteinSequencingCNNModel.h5"
#         self.model_batchsize = 64
#         self.model_epochs = 1
#         self.model = Sequential()
#
#     def createModel(self):
#         # https://keras.io/getting-started/sequential-model-guide/
#         # Reset model
#
#         model_params = [1, 32, 1, 1]
#
#         self.model = Sequential()
#
#         # Layer 1
#         # From TF Conv2DFunction: input_shape = (128, 128, 3) for 128x128 RGB pictures in `data_format where "channels_last"
#
#         self.model.add(Conv2D(32, (1, 3), activation='relu',
#                               input_shape=(self.input_Height, self.input_Width, self.input_Channels)))  #
#         self.model.add(Conv2D(32, (1, 3), activation='relu'))
#         self.model.add(MaxPooling2D(pool_size=(1, 1)))
#         self.model.add(Dropout(0.25))
#
#         # Layer 2
#         self.model.add(Conv2D(64, (1, 1), activation='relu'))
#         self.model.add(Conv2D(64, (1, 1), activation='relu'))
#         self.model.add(MaxPooling2D(pool_size=(1, 1)))
#         self.model.add(Dropout(0.25))
#
#         # Layer 3
#         self.model.add(Flatten())
#         self.model.add(Dense(256, activation='relu'))
#         self.model.add(Dropout(0.5))
#         self.model.add(Dense(4, activation='softmax'))
#
#         # Layer 4
#         sgd = SGD(lr=model_learningRate, decay=1e-6, momentum=0.9, nesterov=True)
#         self.model.compile(loss='categorical_crossentropy', optimizer=sgd)
#
#         self.model.compile(optimizer='rmsprop',
#                            loss='categorical_crossentropy',
#                            metrics=['accuracy'])
#
#         return self.model
#
#
# # https://missinglink.ai/guides/convolutional-neural-networks/convolutional-neural-network-tutorial-basic-advanced/
# model.add(Conv2D(32, kernel_size=(5, 5), strides=(1, 1),
#                  activation='relu',
#                  input_shape=input_shape))
# model.add(MaxPooling2D(pool_size=(2, 2), strides=(2, 2)))
# model.add(Conv2D(64, (5, 5), activation='relu'))
# model.add(MaxPooling2D(pool_size=(2, 2)))
# model.add(Flatten())
# model.add(Dense(1000, activation='relu'))
# model.add(Dense(num_classes, activation='softmax'))
#
#
# layer1 = create_new_conv_layer(x_shaped, 1, 32, [5, 5], [2, 2], name='layer1')
# layer2 = create_new_conv_layer(layer1, 32, 64, [5, 5], [2, 2], name='layer2')
# flattened = tf.reshape(layer2, [-1, 7 * 7 * 64])



# -*- coding: utf-8 -*-
"""
################################################################################################

This file contains functions to get protein sequences.

To download a protein sequence from the uniprot (http://www.uniprot.org/) website,
just provide a protein ID or a txt file with uniprot ID, obtaining a txt file with the protein sequences.

Authors: Ana Marta Sequeira

Date:

Email:

################################################################################################
"""

import urllib.request, urllib.parse, urllib.error
import string


def get_protein_sequence(protein_id):
    """
    Get the protein sequence from the uniprot website by ID.

    :param protein_id: ProteinID is a string indicating ID such as "P48039".
    :return: str protein sequence
    """

    ID=str(protein_id)
    url='http://www.uniprot.org/uniprot/'+ID+'.fasta'
    #print(url)
    localfile=urllib.request.urlopen(url)

    temp=localfile.readlines()
    res=''
    for i in range(1,len(temp)):
        res=res+str.strip(str(temp[i]))
    AALetter=["A","R","N","D","C","Q","E","G","H","I","L","K","M","F","P","S","T","W","Y","V"]
    proteinsequence=''.join(c for c in res if c in AALetter)

    return proteinsequence


def get_protein_sequence_from_txt(path, openfile, savefile):
    """
    Get the protein sequence from the uniprot website by the file containing ID.

    :param path: path is a directory path containing the ID file (e.g. "/home/protein/")
    :param openfile: ID file  (e.g. "proteinID.txt")
    :param savefile: file saving the obtained protein sequences (e.g. "protein.txt")
    :return: file txt with protein sequences
    """

    f1=open(path+savefile,'wb')
    f2=open(path+openfile,'r')
    #	res=[]
    for index,i in enumerate(f2):

        itrim=str.strip(i)
        if itrim == "":
            continue
        else:
            temp=get_protein_sequence(itrim)
            print("--------------------------------------------------------")
            print("The %d protein sequence has been downloaded!" %(index+1))
            print(temp)
            f1.write(str.encode(temp+'\n'))
            print("--------------------------------------------------------")
    #		res.append(temp+'\n')
    #	f1.writelines(res)
    f2.close()
    f1.close()
    return 'end'


if __name__ == '__main__':

    uniprot=get_protein_sequence('P08172')
    print(uniprot)

    result_ID=get_protein_sequence_from_txt(r"", "data/target.txt", "data/result.txt")
    print(result_ID)









"""
##############################################################################

File with useful functions to plot curves to better understanding of ML and DL models (udes in these classes)
Authors: Ana Marta Sequeira

Date:12/2020

Email:

##############################################################################
"""

import sys

import seaborn as sns
import numpy as np
import matplotlib.pyplot as plt
import warnings
from numpy import interp
from itertools import cycle
from sklearn.metrics import *
from sklearn.model_selection import learning_curve
from sklearn.model_selection import ShuffleSplit, train_test_split
from sklearn.model_selection import validation_curve
from sklearn.metrics import make_scorer
from sklearn.metrics import matthews_corrcoef
from sklearn.metrics import roc_curve, auc
from sklearn.preprocessing import label_binarize
from sklearn.multiclass import OneVsRestClassifier



warnings.filterwarnings("ignore")
# Seed for randomization. Set to some definite integer for debugging and set to None for production
seed = None


def plot_summary_accuracy(model, path_save=None, show=True):
    """
    Function to plot training and validation accuracy
    :param model:
    :param path_save:
    :param show:
    :return:
    """
    dnn = model
    plt.clf()
    # summarize history for accuracy
    plt.plot(dnn.history['accuracy'])
    plt.plot(dnn.history['val_accuracy'])
    plt.title('model accuracy')
    plt.ylabel('accuracy')
    plt.xlabel('epoch')
    plt.legend(['train', 'validation'], loc='upper left')
    if path_save is not None:
        plt.savefig(fname=path_save)
    if show is True:
        plt.show(block=True)
    plt.clf()


def plot_summary_loss(model, path_save=None, show=True):
    """
    Function to plot validationa nd training loss
    :param model:
    :param path_save:
    :param show:
    :return:
    """
    dnn = model
    # summarize history for loss
    plt.clf()
    plt.plot(dnn.history['loss'])
    plt.plot(dnn.history['val_loss'])
    plt.title('model loss')
    plt.ylabel('loss')
    plt.xlabel('epoch')
    plt.legend(['train', 'validation'], loc='upper left')
    if path_save is not None:
        plt.savefig(fname=path_save)
    if show is True:
        plt.show(block=True)
    plt.clf()

def plot_roc_curve(classifier, final_units, x_test, y_test, x_train, y_train, ylim=(0.0, 1.00), xlim=(0.0, 1.0),
                   title='Receiver operating characteristic (ROC) curve',
                   path_save='plot_roc_curve', show=False, batch_size=None):
    """
    Function to plot a ROC curve
     On the y axis, true positive rate and false positive rate on the X axis.
     The top left corner of the plot is the 'ideal' point - a false positive rate of zero, and a true positive rate of one,
     meaning a larger area under the curve (AUC) is usually better.
    :param classifier: {classifier instance} pre-trained classifier used for predictions.
    :param x_test: {array} descriptor values of the tests data.
    :param y_test:{array} true class values of the tests data.
    :param ylim: y-axis limits
    :param xlim: x- axis limits
    :param title: title of plot
    :return: plot ROC curve
    Needs classifier with probability
    https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc.html
    """
    lw = 2
    # binary
    if final_units < 2:
        y_score = classifier.predict(x_test)
        # y_score = classifier.predict_proba(X_test)[:,1]
        fpr, tpr, thresholds = roc_curve(y_test, y_score)
        roc_auc = auc(fpr, tpr)
        print(roc_auc)
        plt.show()
        plt.figure()
        plt.plot(fpr, tpr, color='darkorange',
                 lw=lw, label='ROC curve (area = %0.2f)' % roc_auc)
        plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
        plt.xlim(xlim)
        plt.ylim(ylim)
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title(title)
        plt.legend(loc="lower right", )
        if path_save is not None:
            plt.savefig(fname=path_save)
        if show is True:
            plt.show()
        plt.clf()

    # multiclass/ multilabel
    elif final_units > 2:
        # Binarize the output
        classe = np.unique(y_train)
        y_train = label_binarize(y_train, classes=classe)
        n_classes = y_train.shape[1]
        y_test = label_binarize(y_test, classes=classe)

        # Learn to predict each class against the other
        estimator = OneVsRestClassifier(classifier)
        y_score = estimator.fit(x_train, y_train).predict_proba(x_test)

        # Compute ROC curve and ROC area for each class
        fpr = dict()
        tpr = dict()
        roc_auc = dict()
        for i in range(n_classes):
            fpr[i], tpr[i], _ = roc_curve(y_test[:, i], y_score[:, i])
            roc_auc[i] = auc(fpr[i], tpr[i])

        # Compute micro-average ROC curve and ROC area
        fpr["micro"], tpr["micro"], _ = roc_curve(y_test.ravel(), y_score.ravel())
        roc_auc["micro"] = auc(fpr["micro"], tpr["micro"])

        # First aggregate all false positive rates
        all_fpr = np.unique(np.concatenate([fpr[i] for i in range(n_classes)]))

        # Then interpolate all ROC curves at this points
        mean_tpr = np.zeros_like(all_fpr)
        for i in range(n_classes):
            mean_tpr += interp(all_fpr, fpr[i], tpr[i])

        # Finally average it and compute AUC
        mean_tpr /= n_classes

        fpr["macro"] = all_fpr
        tpr["macro"] = mean_tpr
        roc_auc["macro"] = auc(fpr["macro"], tpr["macro"])

        # Plot all ROC curves
        plt.figure()
        plt.plot(fpr["micro"], tpr["micro"],
                 label='micro-average ROC curve (area = {0:0.2f})'
                       ''.format(roc_auc["micro"]),
                 color='deeppink', linestyle=':', linewidth=4)

        plt.plot(fpr["macro"], tpr["macro"],
                 label='macro-average ROC curve (area = {0:0.2f})'
                       ''.format(roc_auc["macro"]),
                 color='navy', linestyle=':', linewidth=4)

        # Get Unique ec
        color_labels = np.unique(y_train)

        # List of colors in the color palettes
        rgb_values = sns.color_palette("nipy_spectral", final_units)  # 'set2'
        # Map ec to the colors
        for i, color in zip(range(final_units), rgb_values):
            plt.plot(fpr[i], tpr[i], color=color, lw=lw,
                     label='ROC curve of class {0} (area = {1:0.2f})'
                           ''.format(i, roc_auc[i]))

        plt.plot([0, 1], [0, 1], 'k--', lw=lw)
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title(title)
        plt.legend(loc="lower right", fontsize='xx-small')
        if path_save is not None:
            plt.savefig(fname=path_save)
        if show is True:
            plt.show()
        plt.clf()


def plot_precision_recall_curve(y, y_pred, n_classes, show=False, path_save=''):
    """
    Performs plot precision-recall curve
    :param y: Series or Dataframe of true labels
    :param y_pred: Series or Dataframe of predicted labels
    :return: None
    """

    if n_classes < 2:
        plt.clf()
        plt.figure()
        precision, recall, thresholds = precision_recall_curve(y, y_pred)
        plt.plot(recall, precision)
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.0])
        plt.xlabel("Recall")
        plt.ylabel("Precision")
        plt.title("Precision-Recall Curve")

    else:
        # For each class
        # tem q estar ONE vs ONE classifier o y_pred tem de estar com os para shallow machine learning
        precision = dict()
        recall = dict()
        average_precision = dict()
        for i in range(n_classes):
            precision[i], recall[i], _ = precision_recall_curve(y[:, i], y_pred[:, i])
            average_precision[i] = average_precision_score(y[:, i], y_pred[:, i])
        # A "micro-average": quantifying score on all classes jointly
        precision["micro"], recall["micro"], _ = precision_recall_curve(y.ravel(),
                                                                        y_pred.ravel())
        average_precision["micro"] = average_precision_score(y, y_pred, average="micro")
        print('Average precision score, micro-averaged over all classes: {0:0.2f}'.format(average_precision["micro"]))

        # setup plot details
        colors = cycle(['navy', 'turquoise', 'darkorange', 'cornflowerblue', 'teal'])
        plt.figure(figsize=(7, 8))
        f_scores = np.linspace(0.2, 0.8, num=int(n_classes + 1))
        lines = []
        labels = []
        for f_score in f_scores:
            x = np.linspace(0.01, 1)
            y = f_score * x / (2 * x - f_score)
            l, = plt.plot(x[y >= 0], y[y >= 0], color='gray', alpha=0.2)
            plt.annotate('f1={0:0.1f}'.format(f_score), xy=(0.9, y[45] + 0.02))

        lines.append(l)
        labels.append('iso-f1 curves')
        l, = plt.plot(recall["micro"], precision["micro"], color='gold', lw=2)
        lines.append(l)
        labels.append('micro-average Precision-recall (area = {0:0.2f})'
                      ''.format(average_precision["micro"]))

        for i, color in zip(range(n_classes), colors):
            l, = plt.plot(recall[i], precision[i], color=color, lw=2)
            lines.append(l)
            labels.append('Precision-recall for class {0} (area = {1:0.2f})'
                          ''.format(i, average_precision[i]))

        fig = plt.gcf()
        fig.subplots_adjust(bottom=0.25)
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('Recall')
        plt.ylabel('Precision')
        plt.title('Extension of Precision-Recall curve to multi-class')
        plt.legend(lines, labels, loc=(0, -.38), prop=dict(size=14))

    if path_save is not None:
        plt.savefig(fname=path_save)
    if show is True:
        plt.show()
    plt.clf()


def plot_validation_curve(classifier, x_train, y_train, param_name, param_range,
                          cv=5,
                          score=make_scorer(matthews_corrcoef), title="Validation Curve",
                          xlab="parameter range", ylab="MCC", n_jobs=1, show=False, path_save='plot_validation_curve',
                          **params):
    """This function plots a cross-validation curve for the specified classifier on all tested parameters given in the
    option 'param_range'.

    :param classifier: {classifier instance} classifier or validation curve (e.g. sklearn.svm.SVC).
    :param x_train: {array} descriptor values for training data.
    :param y_train: {array} class values for training data.
    :param param_name: {string} parameter to assess in the validation curve plot. For example,
    For SVM,
        "clf__C" (C parameter), "clf__gamma" (gamma parameter).
    For Random Forest,
        "clf__n_estimators" (number of trees),"clf__max_depth" (max num of branches per tree, "clf__min_samples_split" (min number of samples required to
        split an internal tree node), "clf__min_samples_leaf" (min number of samples in newly created leaf).
    :param param_range: {list} parameter range for the validation curve.
    :param cv: {int} number of folds for cross-validation.
    :param score: {metrics instance} scoring function built from make_scorer() or a predefined value in string form
        `sklearn.model_evaluation.scoring-parameter
        <http://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter>`_.
    :param title: {str} graph title
    :param xlab: {str} x axis label.
    :param ylab: {str} y axis label.
    :param n_jobs: {int} number of parallel jobs to use for calculation. if ``-1``, all available cores are used.
    :param filename: {str} if filename given the figure is stored in the specified path.
    :return: plot of the validation curve.

    Based on a function from moodlamp
    Müller A. T. et al. (2017) modlAMP: Python for anitmicrobial peptides, Bioinformatics 33, (17), 2753-2755,
    DOI:10.1093/bioinformatics/btx285
    """
    # just goes with numeric parameters not with string or tuple ones. problem in the axis the plotting semilogx and fill between, but actually does notmake sense plot not numeric values
    train_scores, test_scores = validation_curve(classifier, x_train, y_train, param_name, param_range,
                                                 cv=cv, scoring=score, n_jobs=n_jobs, **params)

    train_mean = np.mean(train_scores, axis=1)
    train_std = np.std(train_scores, axis=1)
    test_mean = np.mean(test_scores, axis=1)
    test_std = np.std(test_scores, axis=1)

    # plotting
    plt.clf()
    plt.title(title)
    plt.xlabel(xlab)
    plt.ylabel(ylab)
    plt.ylim(0.0, 1.1)

    plt.semilogx(param_range, train_mean, label="Training score", color="b")
    plt.fill_between(param_range, train_mean - train_std, train_mean + train_std, alpha=0.2, color="b")
    plt.semilogx(param_range, test_mean, label="Cross-validation score", color="g")
    plt.fill_between(param_range, test_mean - test_std, test_mean + test_std, alpha=0.2, color="g")
    plt.legend(loc="best")
    if path_save is not None:
        plt.savefig(fname=path_save)
    if show is True:
        plt.show()
    plt.clf()


def plot_learning_curve(classifier, x_train, y_train, title='Learning curve', ylim=None,
                        cv=ShuffleSplit(n_splits=10, test_size=0.2, random_state=0),
                        n_jobs=1, train_sizes=np.linspace(.1, 1.0, 5),
                        path_save='plot_learning_curve', show=False, scalability=True, performance=True):
    """
    Plot a learning curve to determine cross validated training and tests scores for different training set sizes
    :param estimator: classifier/ model to use
    :param title: title of the plot
    :param ylim:
    :param cv: cross validation to use
    :param n_jobs:  number of parallel jobs to use for calculation. if ``-1``, all available cores are used.
    :param train_sizes: train sizes to tests
    :return: graphic representing learning curves, numbers of trainig examples, scores on training sets, and scores on tests set
    https://scikit-learn.org/stable/auto_examples/model_selection/plot_learning_curve.html#sphx-glr-auto-examples-model-selection-plot-learning-curve-py
    """

    train_sizes, train_scores, test_scores, fit_times, _ = learning_curve(classifier, x_train, y_train, cv=cv,
                                                                          n_jobs=n_jobs, train_sizes=train_sizes,
                                                                          return_times=True)
    train_scores_mean = np.mean(train_scores, axis=1)
    train_scores_std = np.std(train_scores, axis=1)
    test_scores_mean = np.mean(test_scores, axis=1)
    test_scores_std = np.std(test_scores, axis=1)
    fit_times_mean = np.mean(fit_times, axis=1)
    fit_times_std = np.std(fit_times, axis=1)

    # learning curve
    plt.clf()
    plt.figure()
    plt.grid(b=True, which='both')
    plt.title(title)
    if ylim is not None:
        plt.ylim(*ylim)
    plt.xlabel("Training examples")
    plt.ylabel("Score")
    plt.fill_between(train_sizes, train_scores_mean - train_scores_std,
                     train_scores_mean + train_scores_std, alpha=0.1,
                     color="r")
    plt.fill_between(train_sizes, test_scores_mean - test_scores_std,
                     test_scores_mean + test_scores_std, alpha=0.1, color="g")
    plt.plot(train_sizes, train_scores_mean, 'o-', color="r",
             label="Training score")
    plt.plot(train_sizes, test_scores_mean, 'o-', color="g",
             label="Cross-validation score")

    plt.legend(loc="best")
    if path_save is not None:
        plt.savefig(fname=str(path_save + 'learning_curve.png'))
    if show is True:
        plt.show()
    plt.clf()

    if scalability:
        # Plot n_samples vs fit_times
        plt.grid(b=True, which='both')
        plt.plot(train_sizes, fit_times_mean, 'o-')
        plt.fill_between(train_sizes, fit_times_mean - fit_times_std,
                         fit_times_mean + fit_times_std, alpha=0.1)
        plt.xlabel("Training examples")
        plt.ylabel("fit_times")
        plt.title("Scalability of the model")
        if path_save is not None:
            plt.savefig(fname=str(path_save + 'scalability_model.png'))
        if show is True:
            plt.show()
        plt.clf()

    if performance:
        # Plot fit_time vs score
        plt.grid(b=True, which='both')
        plt.plot(fit_times_mean, test_scores_mean, 'o-')
        plt.fill_between(fit_times_mean, test_scores_mean - test_scores_std,
                         test_scores_mean + test_scores_std, alpha=0.1)
        plt.xlabel("fit_times")
        plt.ylabel("Score")
        plt.title("Performance of the model")
        if path_save is not None:
            plt.savefig(fname=str(path_save + 'performance_model.png'))
        if show is True:
            plt.show()
        plt.clf()

    # todo check if it is better way to do this. option of the plots. arrange the model names. in ones it gets the model names and the others dont
    return

    #
    #
    #
    # @staticmethod
    # def precision_recall_curve_binary(y, y_pred):
    #     """
    #     Performs plot precision-recall curve
    #     :param y: Series or Dataframe of true labels
    #     :param y_pred: Series or Dataframe of predicted labels
    #     :return: None
    #     """
    #     plt.figure()
    #     precision, recall, thresholds = precision_recall_curve(y, y_pred)
    #     plt.plot(recall, precision)
    #     plt.xlim([0.0, 1.0])
    #     plt.ylim([0.0, 1.0])
    #     plt.xlabel("Recall")
    #     plt.ylabel("Precision")
    #     plt.title("Precision-Recall Curve")
    #     plt.show()
    #     plt.close()
    #
    # @staticmethod
    # def precision_recall_curve_multiclass(y, y_pred, n_classes):
    #     # For each class
    #     # tem q estar ONE vs ONE classifier o y_pred tem de estar com os paara shallow machine learning
    #     precision = dict()
    #     recall = dict()
    #     average_precision = dict()
    #     for i in range(n_classes):
    #         precision[i], recall[i], _ = precision_recall_curve(y[:, i], y_pred[:, i])
    #         average_precision[i] = average_precision_score(y[:, i], y_pred[:, i])
    #     # A "micro-average": quantifying score on all classes jointly
    #     precision["micro"], recall["micro"], _ = precision_recall_curve(y.ravel(),
    #                                                                     y_pred.ravel())
    #     average_precision["micro"] = average_precision_score(y, y_pred, average="micro")
    #     print('Average precision score, micro-averaged over all classes: {0:0.2f}'.format(average_precision["micro"]))
    #
    #     from itertools import cycle
    #     # setup plot details
    #     colors = cycle(['navy', 'turquoise', 'darkorange', 'cornflowerblue', 'teal'])
    #     plt.figure(figsize=(7, 8))
    #     f_scores = np.linspace(0.2, 0.8, num=int(n_classes + 1))
    #     lines = []
    #     labels = []
    #     for f_score in f_scores:
    #         x = np.linspace(0.01, 1)
    #         y = f_score * x / (2 * x - f_score)
    #         l, = plt.plot(x[y >= 0], y[y >= 0], color='gray', alpha=0.2)
    #         plt.annotate('f1={0:0.1f}'.format(f_score), xy=(0.9, y[45] + 0.02))
    #
    #     lines.append(l)
    #     labels.append('iso-f1 curves')
    #     l, = plt.plot(recall["micro"], precision["micro"], color='gold', lw=2)
    #     lines.append(l)
    #     labels.append('micro-average Precision-recall (area = {0:0.2f})'
    #                   ''.format(average_precision["micro"]))
    #
    #     for i, color in zip(range(n_classes), colors):
    #         l, = plt.plot(recall[i], precision[i], color=color, lw=2)
    #         lines.append(l)
    #         labels.append('Precision-recall for class {0} (area = {1:0.2f})'
    #                       ''.format(i, average_precision[i]))
    #
    #     fig = plt.gcf()
    #     fig.subplots_adjust(bottom=0.25)
    #     plt.xlim([0.0, 1.0])
    #     plt.ylim([0.0, 1.05])
    #     plt.xlabel('Recall')
    #     plt.ylabel('Precision')
    #     plt.title('Extension of Precision-Recall curve to multi-class')
    #     plt.legend(lines, labels, loc=(0, -.38), prop=dict(size=14))
    #     plt.show()



# -*- coding: utf-8 -*-
"""
##############################################################################

File containing a class to do manifold learning.
Manifold learning is an approach to non-linear dimensionality reduction.
This class performs t-sne and UMAP.
For additional information please read:
    https://umap-learn.readthedocs.io/en/latest/
    https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html
Authors: Ana Marta Sequeira

Date: 12/2020

Email:

##############################################################################
"""
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import umap
import umap.plot
from sklearn.manifold import TSNE
from .ml.utils import timer
sns.set()


class Manifold:
    """
    Class to run manifold learning methods. Contrary to methods such as PCA, manifold preserve nonlinear relationships
    in the data. This class allows to explore t-sne and UMAP: tool to visualize high-dimensional data.
    The class opens with x_data, a dataset containing the features or encodings. Classes of this representations may
    be given in the classes argument (optional). If wanted, already porjected/ embedding matrices
    may be given just to run plots.
    """
    def __init__(self, x_data, classes=None, X_embedded=None, projected = None):
        self.fps_x = x_data # array
        self.X_embedded = X_embedded
        self.classes = classes
        self.mapper = None
        self.umapembedding = None
        self.projected = projected
        self.labels = None

    @timer
    def run_tsne(self, n_components = 2, **params):
        """
        t-sne is tool to visualize high-dimensional data. It converts similarities between data points to joint
        probabilities and tries to minimize the Kullback-Leibler divergence between the joint probabilities of the
        low-dimensional embedding and the high-dimensional data. t-SNE has a cost function that is not convex,
        i.e. with different initializations we can get different results.
        :param n_components: number components to projct t-sne. by default 2.
        :param **params: paameters for scikit learn t-sne
        :return:
        """
        X_embedded = TSNE(n_components= n_components,**params).fit_transform(self.fps_x)
        self.X_embedded = X_embedded
        self.projected = self.X_embedded
        self.embedding = X_embedded
        self.labels ='TSNE'
        return X_embedded

    @timer
    def run_umap(self,n_neighbors=20, min_dist=0.1, n_components=50,metric='correlation', target=None):
        """
        Function to run UMAP
        :param n_neighbors: how UMAP balances local versus global structure in the data
        low values of n_neighbors will force UMAP to concentrate on very local structure
        (potentially to the detriment of the big picture), while large values will push UMAP to look at larger
        neighborhoods of each point when estimating the manifold structure of the data, losing fine detail structure
        for the sake of getting the broader of the data. 20 by default
        :param min_dist: how tightly UMAP is allowed to pack points together. It, quite literally,
        provides the minimum distance apart that points are allowed to be in the low dimensional representation.
        This means that low values of min_dist will result in clumpier embeddings. This can be useful if you are
        interested in clustering, or in finer topological structure. Larger values of min_dist will prevent UMAP
        from packing points together and will focus on the preservation of the broad topological structure instead.
        0.1 by default
        :param n_components: components to consider. 50 by default.
        :param metric: default 'correlation'
        :param target: optional.
        More information on https://umap-learn.readthedocs.io/en/latest/parameters.html#n-neighbors
        :return:
        """

        reducer = umap.UMAP(n_neighbors = n_neighbors , min_dist=min_dist, n_components=n_components,
                            metric=metric)
        mapper = reducer.fit(self.fps_x)
        embedding = reducer.fit_transform(self.fps_x, y=target)
        self.projected = mapper.embedding_
        self.mapper=mapper
        self.embedding = embedding
        self.labels ='UMAP'
        return mapper, embedding

    def manifold_scatter_plot(self, target, dim1=0, dim2=1, title=None, show=True, path_save='manifold_scatter_plot.png'):
        """
        Function to retrieve scatter plot for 2 dimensions for both t-SNE and UMAP analysis.

        :param target:  list/array of targets of data.
        :param dim1: axis to consider to plot. by default the first: 0
        :param dim2: axis to consider to plot. by default the first: 1
        :param title: title of the generated plot . None by default
        :param show: If t display or not the scatter plot.
        :param path_save: path to save scatter plot. 'manifold_scatter_plot.png' by default
        :return: scatter plot of two components derived from manifold analysis, colored by target.
        """
        plt.clf()

        n_color=len(np.unique(target))
        # Get Unique ec
        color_labels = np.unique(target)

        # List of colors in the color palettes
        rgb_values = sns.color_palette("nipy_spectral", n_color) # 'set2'

        # Map ec to the colors
        color_map = dict(zip(color_labels, rgb_values))

        # Finally use the mapped values
        plt.scatter(self.projected[:, dim1], self.projected[:, dim2],
                    c=target.map(color_map), s=0.3)

        # create legend
        # classes  - colour labels
        # class_colours rgb values
        recs = []
        from matplotlib.lines import Line2D

        for i in range(0, n_color):
            recs.append(Line2D((0,0.75),(0,0), color=rgb_values[i], marker='o', linestyle=''))
            # recs.append(mpatches.Circle((0, 0), 0.5, fc=rgb_values[i]))

        n_col = int(n_color/23)+1

        lgd = plt.legend(recs, color_labels, bbox_to_anchor=(1,1), loc="upper left",
                         fontsize='xx-small', ncol = n_col, shadow=False, title='classes')
        # add labels and title
        plt.xlabel('dim 1')
        plt.ylabel('dim 2')
        if title is None:
            title='{} scatter plot of 2 dimensions'.format(self.labels)
        plt.title(title)

        # save and empty plt
        if path_save is not None:
            plt.savefig(fname=path_save,bbox_extra_artists=(lgd,), bbox_inches='tight', pad_inches=0)
        if show is True:
            plt.show()
        plt.clf()

    # def manifold_scatter_plot2(self, target, title=None, show=True, path_save='manifold_scatter_plot.png'):
    #     fig, ax = plt.subplots(1, figsize=(14, 10))
    #     plt.scatter(self.projected[:, 0], self.projected[:, 1], s=0.3, c=target, cmap='Spectral', alpha=1.0)
    #     plt.setp(ax, xticks=[], yticks=[])
    #     cbar = plt.colorbar(boundaries=np.arange(11)-0.5)
    #     cbar.set_ticks(np.arange(len(np.unique(target))))
    #     cbar.set_ticklabels(np.unique(target))
    #     plt.title('Fashion MNIST Embedded via UMAP')
    #     if title is None:
    #         title='{} scatter plot of 2 dimensions'.format(self.labels)
    #     plt.title(title)
    #
    #     # save and empty plt
    #     if path_save is not None:
    #         plt.savefig(fname=path_save, bbox_inches='tight', pad_inches=0)
    #     if show is True:
    #         plt.show()
    #     plt.clf()






# https://arxiv.org/pdf/1802.03426.pdf
# https://umap-learn.readthedocs.io/en/latest/parameters.html


 # much data perplixity between 5-50 perplixity , very large try 100
# https://blog.clairvoyantsoft.com/mlmuse-visualisation-of-high-dimensional-data-using-t-sne-ac6264316d7f
# https://distill.pub/2016/misread-tsne/
# use pca or svd before of data to high dimensional
# https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html
# https://stats.stackexchange.com/questions/238538/are-there-cases-where-pca-is-more-suitable-than-t-sne
# plot embeddings / word embeddings
# https://towardsdatascience.com/visualising-high-dimensional-datasets-using-pca-and-t-sne-in-python-8ef87e7915b



# to visualize embeddings
# https://umap-learn.readthedocs.io/en/latest/embedding_space.html in space 2 D or for example i a sphere and can in jupyter interactive maps
# https://www.kaggle.com/colinmorris/visualizing-embeddings-with-t-sne

